#  This code is a part of X-ray: Generate and Analyse (XGA), a module designed for the XMM Cluster Survey (XCS).
#  Last modified by David J Turner (turne540@msu.edu) 20/02/2023, 14:04. Copyright (c) The Contributors


from typing import Tuple, List, Union

import numpy as np
from astropy.units import Quantity, pix, deg, UnitConversionError, UnitBase, Unit
from astropy.wcs import WCS

from ..products import Image, RateMap, ExpMap
from ..sourcetools import ang_to_rad, rad_to_ang
from ..utils import xmm_sky


def pix_deg_scale(coord: Quantity, input_wcs: WCS, small_offset: Quantity = Quantity(1, 'arcmin')) -> Quantity:
    """
    Very heavily inspired by the regions module version of this function, just tweaked to work better for
    my use case. Perturbs the given coordinates with the small_offset value, converts the changed ra-dec
    coordinates to pixel, then calculates the difference between the new and original coordinates in pixel.
    Then small_offset is converted to degrees and  divided by the pixel distance to calculate a pixel to degree
    factor.

    :param Quantity coord: The starting coordinates.
    :param WCS input_wcs: The world coordinate system used to calculate the pixel to degree scale
    :param Quantity small_offset: The amount you wish to perturb the original coordinates
    :return: Factor that can be used to convert pixel distances to degree distances, returned as an astropy
        quantity with units of deg/pix.
    :rtype: Quantity
    """
    if coord.unit != pix and coord.unit != deg:
        raise UnitConversionError("This function can only be used with radec or pixel coordinates as input")
    elif coord.shape != (2,):
        raise ValueError("coord input must only contain 1 pair.")
    elif not small_offset.unit.is_equivalent("deg"):
        raise UnitConversionError("small_offset must be convertible to degrees")

    if coord.unit == deg:
        pix_coord = Quantity(input_wcs.all_world2pix(*coord.value, 0), pix)
        deg_coord = coord
    elif coord.unit == pix:
        deg_coord = Quantity(input_wcs.all_pix2world(*coord.value, 0), deg)
        pix_coord = coord
    else:
        raise UnitConversionError('{} is not a recognised position unit'.format(coord.unit.to_string()))

    perturbed_coord = deg_coord + Quantity([0, small_offset.to("deg").value], 'deg')
    perturbed_pix_coord = Quantity(input_wcs.all_world2pix(*perturbed_coord.value, 0), pix)

    diff = abs(perturbed_pix_coord - pix_coord)
    pix_dist = np.hypot(*diff)

    scale = small_offset.to('deg').value / pix_dist.value

    return Quantity(scale, 'deg/pix')


def sky_deg_scale(im_prod: Union[Image, RateMap, ExpMap], coord: Quantity,
                  small_offset: Quantity = Quantity(1, 'arcmin')) -> Quantity:
    """
    This is equivelant to pix_deg_scale, but instead calculates the conversion factor between
    XMM's XY sky coordinate system and degrees.

    :param Image/Ratemap/ExpMap im_prod: The image product to calculate the conversion factor for.
    :param Quantity coord: The starting coordinates.
    :param Quantity small_offset: The amount you wish to perturb the original coordinates
    :return: A scaling factor to convert sky distances to degree distances, returned as an astropy
        quantity with units of deg/xmm_sky.
    :rtype: Quantity
    """
    # Now really this function probably isn't necessary at, because there is a fixed scaling from degrees
    #  to this coordinate system, but I do like to be general

    if coord.shape != (2,):
        raise ValueError("coord input must only contain 1 pair.")
    elif not small_offset.unit.is_equivalent("deg"):
        raise UnitConversionError("small_offset must be convertible to degrees")

    # Seeing as we're taking an image product input on this one, I can leave the checking
    #  of inputs to coord_conv
    # We need the degree and xmm_sky original coordinates
    deg_coord = im_prod.coord_conv(coord, deg)
    sky_coord = im_prod.coord_conv(coord, xmm_sky)

    perturbed_coord = deg_coord + Quantity([0, small_offset.to("deg").value], 'deg')
    perturbed_sky_coord = im_prod.coord_conv(perturbed_coord, xmm_sky)

    diff = abs(perturbed_sky_coord - sky_coord)
    sky_dist = np.hypot(*diff)

    scale = small_offset.to('deg').value / sky_dist.value

    return Quantity(scale, deg/xmm_sky)


def pix_rad_to_physical(im_prod: Union[Image, RateMap, ExpMap], pix_rad: Quantity, out_unit: Union[UnitBase, str],
                        coord: Quantity, z: Union[float, int] = None, cosmo=None) -> Quantity:
    """
    Pure convenience function to convert a list of pixel radii to whatever unit we might want at the end. Used
    quite a lot in the imagetools.profile functions, which is why it was split off into its own function. Redshift
    and cosmology must be supplied if proper distance units (like kpc) are chosen for out_unit.

    :param Image/RateMap/ExpMap im_prod: The image/ratemap product for which the conversion is taking place.
    :param Quantity pix_rad: The array of pixel radii to convert to out_unit.
    :param UnitBase/str out_unit: The desired output unit for the radii, either an astropy unit object or a name string.
    :param Quantity coord: The position of the object being analysed.
    :param float/int z: The redshift of the object (only required for proper distance units like kpc).
    :param cosmo: The chosen cosmology for the analysis (only required for proper distance units like kpc).
    :return: An astropy Quantity with the radii in units of out_unit.
    :rtype: Quantity
    """
    if pix_rad.unit != pix:
        raise UnitConversionError("pix_rads must be in units of pixels")

    # See what type of unit input was given
    if isinstance(out_unit, str):
        out_unit = Unit(out_unit)

    deg_rads = Quantity(pix_deg_scale(coord, im_prod.radec_wcs).value * pix_rad.value, 'deg')

    if out_unit.is_equivalent("kpc") and z is not None and cosmo is not None:
        # Quick convert to kpc with my handy function and then go to whatever unit the user requested
        # Wham bam pixels to proper distance
        conv_rads = ang_to_rad(deg_rads, z, cosmo).to(out_unit)
    elif out_unit.is_equivalent("kpc") and (z is None or cosmo is None):
        raise ValueError("If you wish to convert to physical units such as kpc, you must supply "
                         "a redshift and cosmology")
    elif out_unit.is_equivalent("deg"):
        conv_rads = deg_rads.to(out_unit)
    elif out_unit == pix:
        # Commenting out the sassy warning for now...
        # warn("You're converting pixel radii to pixels...")
        conv_rads = pix_rad
    else:
        conv_rads = None
        raise UnitConversionError("cen_rad_units doesn't appear to be a distance or angular unit.")

    return conv_rads


def physical_rad_to_pix(im_prod: Union[Image, RateMap, ExpMap], physical_rad: Quantity,
                        coord: Quantity, z: Union[float, int] = None, cosmo=None) -> Quantity:
    """
    Another convenience function, this time to convert physical radii to pixels. It can deal with both angular and
    proper radii, so long as redshift and cosmology information is provided for the conversion from proper radii
    to pixels.

    :param Image/RateMap/ExpMap im_prod:
    :param Quantity physical_rad: The physical radius to be converted to pixels.
    :param Quantity coord: The position of the object being analysed.
    :param float/int z: The redshift of the object (only required for input proper distance units like kpc).
    :param cosmo: The chosen cosmology for the analysis (only required for input proper distance units like kpc).
    :return: The converted radii, in an astropy Quantity with pix units.
    :rtype: Quantity
    """

    if physical_rad.unit.is_equivalent("kpc") and z is not None and cosmo is not None:
        conv_rads = rad_to_ang(physical_rad, z, cosmo).to('deg')
    elif physical_rad.unit.is_equivalent("kpc") and (z is None or cosmo is None):
        raise ValueError("If you wish to convert to convert from proper distance units such as kpc, you must supply "
                         "a redshift and cosmology")
    elif physical_rad.unit.is_equivalent("deg"):
        conv_rads = physical_rad.to('deg')
    elif physical_rad.unit == pix:
        raise UnitConversionError("You are trying to convert from pixel units to pixel units.")
    else:
        conv_rads = None
        raise UnitConversionError("cen_rad_units doesn't appear to be a distance or angular unit.")

    phys_to_pix = 1 / pix_deg_scale(coord, im_prod.radec_wcs).value
    conv_rads = Quantity(conv_rads.value * phys_to_pix, 'pix')

    return conv_rads


def data_limits(im_prod: Union[Image, RateMap, ExpMap, np.ndarray]) -> Tuple[List[int], List[int]]:
    """
    A function that finds the pixel coordinates that bound where data is present in
    Image or RateMap object.

    :param Image/RateMap/ndarray im_prod: An Image, RateMap, or numpy array that you wish to find
        boundary coordinates for.
    :return: Two lists, the first with the x lower and upper bounding coordinates, and the second with
        the y lower and upper bounding coordinates.
    :rtype: Tuple[List[int, int], List[int, int]]
    """
    if isinstance(im_prod, Image) and im_prod.data.sum() != 0:
        # For the XGA Image products
        # This just finds out where the zeros in the data are
        locations = np.where(im_prod.data != 0)
    elif isinstance(im_prod, np.ndarray) and im_prod.sum() != 0:
        # For numpy arrays
        locations = np.where(im_prod != 0)
    else:
        raise ValueError("Supplied data only contains zeros, data limits cannot be found in this case.")

    # Finds the maximum and minimum locations of zeros in both x and y spaces - these are the boundary coordinates
    # Adds and subtracts 1 to give a very small border.
    x_min = locations[1].min() - 1
    x_max = locations[1].max() + 1

    y_min = locations[0].min() - 1
    y_max = locations[0].max() + 1

    # Returns the boundary coordinates in pairs of min, max.
    return [x_min, x_max], [y_min, y_max]


def edge_finder(data: Union[RateMap, ExpMap, np.ndarray], keep_corners: bool = True,
                border: bool = False) -> np.ndarray:
    """
    A simple edge finding algorithm designed to locate 'edges' in binary data, or in special cases produce a detector
    map of an instrument using an exposure map. The algorithm takes the difference of one column from the next, over
    the entire array, then does the same with rows. Different difference values indicate where edges are in the array,
    and when added together all edges should be located.

    Depending on how the 'border' option is set, the returned array will either represent the exact edge, or
    a boundary that is 1 pixel outside the actual edge.

    :param RateMap/ExpMap/ndarray data: The 2D array or exposure map to run edge detection on. If an array is
        passed it must only consist of 0s and 1s.
    :param bool keep_corners: Should corner information be kept in the output array. If True then 2s in the
        output will indicate vertices.
    :param bool border: If True, then the returned array will represent a border running around the boundary of the
        true edge, rather than the outer boundary of the edge itself.
    :return: An array of 0s and 1s. 1s indicate a detected edge.
    :rtype: np.ndarray
    """
    if not isinstance(data, (RateMap, ExpMap, np.ndarray)) or (isinstance(data, np.ndarray)
                                                               and np.where((data != 1) & (data != 0))[0].any()):
        raise TypeError("This simple edge-finding algorithm only works on exposure maps (whether passed directly or "
                        "accessed from a RateMap), or arrays of ones and zeros.")
    elif isinstance(data, ExpMap):
        dat_map = data.data.copy()
        # Turn the exposure map into something simpler, either on a detector or not
        dat_map[dat_map != 0] = 1
    elif isinstance(data, RateMap):
        dat_map = data.expmap.data.copy()
        # Turn the exposure map from the ratemap into something simpler, either on a detector or not
        dat_map[dat_map != 0] = 1
    elif isinstance(data, np.ndarray):
        dat_map = data.copy()

    # Do the diff from top to bottom of the image, the append option adds a line of zeros at the end
    #  otherwise the resulting array would only be N-1 elements 'high'.
    hori_edges = np.diff(dat_map, axis=0, append=0)

    if not border:
        # A 1 in this array means you're going from no chip to on chip, which means the coordinate where 1
        # is recorded is offset by 1 from the actual edge of the chip elements of this array.
        need_corr_y, need_corr_x = np.where(hori_edges == 1)
        # So that's why we add one to those y coordinates (as this is the vertical pass of np.diff
        new_y = need_corr_y + 1
        # Then make sure chip edge = 1, and everything else = 0
        hori_edges[need_corr_y, need_corr_x] = 0
        hori_edges[new_y, need_corr_x] = 1
        # -1 in this means going from chip to not-chip
        hori_edges[hori_edges == -1] = 1
    else:
        need_corr_y, need_corr_x = np.where(hori_edges == -1)
        # So that's why we add one to those y coordinates (as this is the vertical pass of np.diff
        new_y = need_corr_y + 1
        # Then make sure chip edge = 1, and everything else = 0
        hori_edges[need_corr_y, need_corr_x] = 0
        hori_edges[new_y, need_corr_x] = 1

    # The same process is repeated here, but in the x direction, so you're finding vertical edges
    vert_edges = np.diff(dat_map, axis=1, append=0)

    if not border:
        need_corr_y, need_corr_x = np.where(vert_edges == 1)
        new_x = need_corr_x + 1
        vert_edges[need_corr_y, need_corr_x] = 0
        vert_edges[need_corr_y, new_x] = 1
        vert_edges[vert_edges == -1] = 1
    else:
        need_corr_y, need_corr_x = np.where(vert_edges == -1)
        new_x = need_corr_x + 1
        vert_edges[need_corr_y, need_corr_x] = 0
        vert_edges[need_corr_y, new_x] = 1

    # Both passes are combined into one, with possible values of 0 (no edge), 1 (edge detected in one pass),
    #  and 2 (edge detected in both pass).
    comb = hori_edges + vert_edges

    # If we don't want to keep information on intersections between two edges (corner values of 2), then we just
    #  set them to one
    if not keep_corners:
        comb[np.where((comb != 1) & (comb != 0))] = 1

    return comb




#  This code is a part of X-ray: Generate and Analyse (XGA), a module designed for the XMM Cluster Survey (XCS).
#  Last modified by David J Turner (turne540@msu.edu) 20/02/2023, 14:04. Copyright (c) The Contributors

import os
import sys

from astropy.units import Quantity

sys.path.append(os.path.abspath("..") + 'xga/')


# Any useful constants
A907_LOC = Quantity([149.59209, -11.05972], 'deg')
im_path = os.path.join(os.path.abspath("."), "test_data/0201903501/images/0201903501_pn_exp1-0.50-2.00keVimg.fits")
exp_path = os.path.join(os.path.abspath("."), "test_data/0201903501/images/0201903501_pn_exp1-0.50-2.00keVexpmap.fits")
A907_IM_PN_INFO = [im_path, '0201903501', 'pn', '', '', '', Quantity(0.5, 'keV'), Quantity(2.0, 'keV')]
A907_EX_PN_INFO = [exp_path, '0201903501', 'pn', '', '', '', Quantity(0.5, 'keV'), Quantity(2.0, 'keV')]




#  This code is a part of X-ray: Generate and Analyse (XGA), a module designed for the XMM Cluster Survey (XCS).
#  Last modified by David J Turner (turne540@msu.edu) 30/07/2024, 17:10. Copyright (c) The Contributors

from typing import Tuple, Union, List
from warnings import warn

import numpy as np
from astropy.units import Quantity

from .deproj import shell_ann_vol_intersect
from .. import NUM_CORES, ABUND_TABLES
from ..exceptions import NoProductAvailableError
from ..imagetools.misc import pix_deg_scale
from ..imagetools.profile import annular_mask
from ..products.profile import GasTemperature3D
from ..samples import BaseSample, ClusterSample
from ..sas import region_setup
from ..sources import BaseSource, GalaxyCluster
from ..xspec.fit import single_temp_apec_profile

ALLOWED_ANN_METHODS = ['min_snr', 'min_cnt']


def _ann_bins_setup(source: BaseSource, outer_rad: Quantity, min_width: Quantity, lo_en: Quantity, hi_en: Quantity,
                    obs_id: str = None, inst: str = None, psf_corr: bool = False, psf_model: str = "ELLBETA",
                    psf_bins: int = 4, psf_algo: str = "rl", psf_iter: int = 15):
    """
    This method just sets up radii, masks, etc. for annular binning functions in this file. The operations in
    this function are shared by multiple other binning functions, hence they have been put in a function of their
    own to minimise duplication.

    :param BaseSource source: The source object to generate annuli for.
    :param Quantity outer_rad: The outermost radius of the source region we will generate annuli within.
    :param Quantity min_width: The minimum allowable width of the annuli. This can be set to try and avoid
        PSF effects.
    :param Quantity lo_en: The lower energy bound of the ratemap to use for the signal to noise calculations.
    :param Quantity hi_en: The upper energy bound of the ratemap to use for the signal to noise calculations.
    :param str obs_id: An ObsID of a specific ratemap to use for the SNR calculations. Default is None, which
            means the combined ratemap will be used. Please note that inst must also be set to use this option.
    :param str inst: The instrument of a specific ratemap to use for the SNR calculations. Default is None, which
        means the combined ratemap will be used.
    :param bool psf_corr: Sets whether you wish to use a PSF corrected ratemap or not.
    :param str psf_model: If the ratemap you want to use is PSF corrected, this is the PSF model used.
    :param int psf_bins: If the ratemap you want to use is PSF corrected, this is the number of PSFs per
        side in the PSF grid.
    :param str psf_algo: If the ratemap you want to use is PSF corrected, this is the algorithm used.
    :param int psf_iter: If the ratemap you want to use is PSF corrected, this is the number of iterations.
    :return: The various variables that this function sets up
    :rtype:
    """
    # Parsing the ObsID and instrument options, see if they want to use a specific ratemap
    if all([obs_id is None, inst is None]):
        # Here the user hasn't set ObsID or instrument, so we use the combined data
        rt = source.get_combined_ratemaps(lo_en, hi_en, psf_corr, psf_model, psf_bins, psf_algo, psf_iter)
        interloper_mask = source.get_interloper_mask()
    elif all([obs_id is not None, inst is not None]):
        # Both ObsID and instrument have been set by the user
        rt = source.get_ratemaps(obs_id, inst, lo_en, hi_en, psf_corr, psf_model, psf_bins, psf_algo, psf_iter)
        interloper_mask = source.get_interloper_mask(obs_id)

    # Just making sure our relevant distances are in the same units, so that we can convert to pixels
    outer_rad = source.convert_radius(outer_rad, 'deg')
    min_width = source.convert_radius(min_width, 'deg')

    # Using the ratemap to get a conversion factor from pixels to degrees, though we will use it
    #  the other way around
    pix_to_deg = pix_deg_scale(source.default_coord, rt.radec_wcs)

    # Making sure to go up to the whole number, pixels have to be integer of course, and I think it's
    #  better to err on the side of caution here and make things slightly wider than requested
    outer_rad = int(np.ceil(outer_rad / pix_to_deg).value)
    min_width = int(np.ceil(min_width / pix_to_deg).value)

    # The maximum possible number of annuli, based on the input outer radius and minimum width
    # We have already made sure that the outer radius and minimum width allowed are integers by using
    #  np.ceil, so we know max_ann is going to be a whole number of annuli
    max_ann = int(outer_rad / min_width)

    # These are the initial bins, with imposed minimum width, I have to add one to max_ann because linspace wants the
    #  total number of values to generate, and while there are max_ann annuli, there are max_ann+1 radial boundaries
    init_rads = np.linspace(0, outer_rad, max_ann + 1).astype(int)
    # Converts the source's default analysis coordinates to pixels
    pix_centre = rt.coord_conv(source.default_coord, 'pix')
    # Sets up a mask to correct for interlopers and weird edge effects
    corr_mask = interloper_mask * rt.edge_mask

    # Setting up our own background region
    back_inn_rad = np.array([np.ceil(source.background_radius_factors[0] * outer_rad)]).astype(int)
    back_out_rad = np.array([np.ceil(source.background_radius_factors[1] * outer_rad)]).astype(int)

    # Using my annular mask function to make a nice background region, which will be corrected for instrumental
    #  stuff and interlopers in a second
    back_mask = annular_mask(pix_centre, back_inn_rad, back_out_rad, rt.shape) * corr_mask

    # Generates the requested annular masks, making sure to apply the correcting mask
    ann_masks = annular_mask(pix_centre, init_rads[:-1], init_rads[1:], rt.shape) * corr_mask[..., None]

    cur_rads = init_rads.copy()

    return rt, cur_rads, max_ann, ann_masks, back_mask, pix_centre, corr_mask, pix_to_deg


def _snr_bins(source: BaseSource, outer_rad: Quantity, min_snr: float, min_width: Quantity, lo_en: Quantity,
              hi_en: Quantity, obs_id: str = None, inst: str = None, psf_corr: bool = False, psf_model: str = "ELLBETA",
              psf_bins: int = 4, psf_algo: str = "rl", psf_iter: int = 15,
              allow_negative: bool = False, exp_corr: bool = True) -> Tuple[Quantity, np.ndarray, int]:
    """
    An internal function that will find the radii required to create annuli with a certain minimum signal to noise
    and minimum annulus width.

    :param BaseSource source: The source object to generate annuli for.
    :param Quantity outer_rad: The outermost radius of the source region we will generate annuli within.
    :param float min_snr: The minimum signal to noise which is allowable in a given annulus.
    :param Quantity min_width: The minimum allowable width of the annuli. This can be set to try and avoid
        PSF effects.
    :param Quantity lo_en: The lower energy bound of the ratemap to use for the signal to noise calculations.
    :param Quantity hi_en: The upper energy bound of the ratemap to use for the signal to noise calculations.
    :param str obs_id: An ObsID of a specific ratemap to use for the SNR calculations. Default is None, which
            means the combined ratemap will be used. Please note that inst must also be set to use this option.
    :param str inst: The instrument of a specific ratemap to use for the SNR calculations. Default is None, which
        means the combined ratemap will be used.
    :param bool psf_corr: Sets whether you wish to use a PSF corrected ratemap or not.
    :param str psf_model: If the ratemap you want to use is PSF corrected, this is the PSF model used.
    :param int psf_bins: If the ratemap you want to use is PSF corrected, this is the number of PSFs per
        side in the PSF grid.
    :param str psf_algo: If the ratemap you want to use is PSF corrected, this is the algorithm used.
    :param int psf_iter: If the ratemap you want to use is PSF corrected, this is the number of iterations.
    :param bool allow_negative: Should pixels in the background subtracted count map be allowed to go below
        zero, which results in a lower signal to noise (and can result in a negative signal to noise).
    :param bool exp_corr: Should signal to noises be measured with exposure time correction, default is True. I
            recommend that this be true for combined observations, as exposure time could change quite dramatically
            across the combined product.
    :return: The radii of the requested annuli, the final snr values, and the original maximum number
        based on min_width.
    :rtype: Tuple[Quantity, np.ndarray, int]
    """

    # This calls a function that just sets things up for this (and other annular binning) function
    rt, cur_rads, max_ann, ann_masks, back_mask, pix_centre, corr_mask, \
        pix_to_deg = _ann_bins_setup(source, outer_rad, min_width, lo_en, hi_en, obs_id, inst, psf_corr, psf_model,
                                     psf_bins, psf_algo, psf_iter)

    if max_ann > 4:
        # This will be modified by the loop until it describes annuli which all have an acceptable signal to noise
        acceptable = False
    else:
        # If there are already 4 or less annuli present then we don't do the reduction while loop, and just take it
        #  as they are, while also issuing a warning
        acceptable = True
        warn("The min_width combined with the outer radius of the source creates only {} initial"
             " annuli, so no re-binning will take place.".format(max_ann), stacklevel=2)
        cur_num_ann = ann_masks.shape[2]
        snrs = []
        for i in range(cur_num_ann):
            # We're calling the signal to noise calculation method of the ratemap for all of our annuli
            snrs.append(rt.signal_to_noise(ann_masks[:, :, i], back_mask, exp_corr, allow_negative))
        # Becomes a numpy array because they're nicer to work with
        snrs = np.array(snrs)

    while not acceptable:
        # How many annuli are there at this point in the loop?
        cur_num_ann = ann_masks.shape[2]

        # Just a list for the snrs to live in
        snrs = []
        for i in range(cur_num_ann):
            # We're calling the signal to noise calculation method of the ratemap for all of our annuli
            snrs.append(rt.signal_to_noise(ann_masks[:, :, i], back_mask, exp_corr, allow_negative))
        # Becomes a numpy array because they're nicer to work with
        snrs = np.array(snrs)
        # We find any indices of the array (== annuli) where the signal to noise is not above our minimum
        bad_snrs = np.where(snrs < min_snr)[0]

        # If there are no annuli below our signal to noise threshold then all is good and joyous and we accept
        #  the current radii
        if len(bad_snrs) == 0:
            acceptable = True
        # We work from the outside of the bad list inwards, and if the outermost bad bin is the one right on the
        #  end of the SNR profile, then we merge that leftwards into the N-1th annuli
        elif len(bad_snrs) != 0 and bad_snrs[-1] == cur_num_ann - 1:
            cur_rads = np.delete(cur_rads, -2)
            ann_masks = annular_mask(pix_centre, cur_rads[:-1], cur_rads[1:], rt.shape) * corr_mask[..., None]
        # A special case must also be added for if the zeroth annulus (i.e. the innermost annulus) isn't meeting the
        #  criteria, because if we leave it to the 'else' statement below then there will be no annulus bound at zero,
        #  which we do require - in this case it means we deleted the 1st annulus
        elif len(bad_snrs) != 0 and bad_snrs[-1] == 0:
            # For where the zeroth annulus is not meeting requirements, we set up this to merge the zeroth and first
            #  annular boundaries
            cur_rads = np.delete(cur_rads, 1)
            ann_masks = annular_mask(pix_centre, cur_rads[:-1], cur_rads[1:], rt.shape) * corr_mask[..., None]
        # Otherwise if the outermost bad annulus is NOT right at the end of the profile, we merge to the right
        else:
            cur_rads = np.delete(cur_rads, bad_snrs[-1])
            ann_masks = annular_mask(pix_centre, cur_rads[:-1], cur_rads[1:], rt.shape) * corr_mask[..., None]

        if ann_masks.shape[2] == 4 and not acceptable:
            warn("The requested annuli for {s} cannot be created, the data quality is too low. As such a set "
                 "of four annuli will be returned".format(s=source.name))
            break

    # Now of course, pixels must become a more useful unit again
    final_rads = (Quantity(cur_rads, 'pix') * pix_to_deg).to("arcsec")

    return final_rads, snrs, max_ann


def _cnt_bins(source: BaseSource, outer_rad: Quantity, min_cnt: Union[int, Quantity],
              min_width: Quantity, lo_en: Quantity, hi_en: Quantity, obs_id: str = None, inst: str = None,
              psf_corr: bool = False, psf_model: str = "ELLBETA", psf_bins: int = 4, psf_algo: str = "rl",
              psf_iter: int = 15) -> Tuple[Quantity, Quantity, int]:
    """
    An internal function that will find the radii required to create annuli with a certain minimum number of counts
    and minimum annulus width.

    :param BaseSource source: The source object to generate annuli for.
    :param Quantity outer_rad: The outermost radius of the source region we will generate annuli within.
    :param float min_cnt: The minimum number of counts which are allowable in a given annulus.
    :param Quantity min_width: The minimum allowable width of the annuli. This can be set to try and avoid
        PSF effects.
    :param Quantity lo_en: The lower energy bound of the ratemap to use for the background subtracted count
        calculations.
    :param Quantity hi_en: The upper energy bound of the ratemap to use for the background subtracted count
        calculations.
    :param str obs_id: An ObsID of a specific ratemap to use for the background subtracted count
        calculations. Default is None, which means the combined ratemap will be used. Please note that inst
        must also be set to use this option.
    :param str inst: The instrument of a specific ratemap to use for the background subtracted count
        calculations. Default is None, which means the combined ratemap will be used.
    :param bool psf_corr: Sets whether you wish to use a PSF corrected ratemap or not.
    :param str psf_model: If the ratemap you want to use is PSF corrected, this is the PSF model used.
    :param int psf_bins: If the ratemap you want to use is PSF corrected, this is the number of PSFs per
        side in the PSF grid.
    :param str psf_algo: If the ratemap you want to use is PSF corrected, this is the algorithm used.
    :param int psf_iter: If the ratemap you want to use is PSF corrected, this is the number of iterations.
    :return: The radii of the requested annuli, the final count values, and the original maximum number
        based on min_width.
    :rtype: Tuple[Quantity, Quantity, int]
    """

    # This just makes sure that the min_cnt variable is the astropy quantity that we expect it to be, otherwise
    #  some of the comparisons made between it and the values returned by background_subtracted_counts will fail
    if type(min_cnt) == int:
        min_cnt = Quantity(min_cnt, 'ct')
    elif (type(min_cnt) == Quantity and not min_cnt.unit.is_equivalent('ct')) or not type(min_cnt) == Quantity:
        raise TypeError("The min_cnt argument must be either an integer, or an astropy Quantity in units of 'ct'.")

    # Run the setup function for these functions that create different annular bins
    rt, cur_rads, max_ann, ann_masks, back_mask, pix_centre, corr_mask, \
        pix_to_deg = _ann_bins_setup(source, outer_rad, min_width, lo_en, hi_en, obs_id, inst, psf_corr, psf_model,
                                     psf_bins, psf_algo, psf_iter)

    if max_ann > 4:
        # This will be modified by the loop until it describes annuli which all have an acceptable signal to noise
        acceptable = False
    else:
        # If there are already 4 or less annuli present then we don't do the reduction while loop, and just take it
        #  as they are, while also issuing a warning
        acceptable = True
        warn("The min_width combined with the outer radius of the source creates only {} initial"
             " annuli, so no re-binning will take place.".format(max_ann), stacklevel=2)
        cur_num_ann = ann_masks.shape[2]
        cnts = []
        for i in range(cur_num_ann):
            # We're calling the background subtracted counts calculation method of the ratemap for all of our annuli
            cnts.append(rt.background_subtracted_counts(ann_masks[:, :, i], back_mask))
        # Becomes an astropy quantity (and so behaves like a numpy array) because they're nicer to work with
        cnts = Quantity(cnts)

    while not acceptable:
        # How many annuli are there at this point in the loop?
        cur_num_ann = ann_masks.shape[2]

        # Just a list for the counts to live in
        cnts = []
        for i in range(cur_num_ann):
            # We're calling the background subtracted count calculation method of the ratemap
            #  for all of our annuli
            cnts.append(rt.background_subtracted_counts(ann_masks[:, :, i], back_mask))
        # Becomes an astropy Quantity (behaves like a numpy array) because they're nicer to work with
        cnts = Quantity(cnts)
        # We find any indices of the array (== annuli) where the counts are not above our minimum
        bad_cnts = np.where(cnts < min_cnt)[0]

        # If there are no annuli below our count threshold then all is good and joyous, and we
        #  accept the current radii
        if len(bad_cnts) == 0:
            acceptable = True
        # We work from the outside of the bad list inwards, and if the outermost bad bin is the one right on the
        #  end of the count profile, then we merge that leftwards into the N-1th annuli
        elif len(bad_cnts) != 0 and bad_cnts[-1] == cur_num_ann - 1:
            cur_rads = np.delete(cur_rads, -2)
            ann_masks = annular_mask(pix_centre, cur_rads[:-1], cur_rads[1:], rt.shape) * corr_mask[..., None]
        # A special case must also be added for if the zeroth annulus (i.e. the innermost annulus) isn't meeting the
        #  criteria, because if we leave it to the 'else' statement below then there will be no annulus bound at zero,
        #  which we do require - in this case it means we deleted the 1st annulus
        elif len(bad_cnts) != 0 and bad_cnts[-1] == 0:
            # For where the zeroth annulus is not meeting requirements, we set up this to merge the zeroth and first
            #  annular boundaries
            cur_rads = np.delete(cur_rads, 1)
            ann_masks = annular_mask(pix_centre, cur_rads[:-1], cur_rads[1:], rt.shape) * corr_mask[..., None]
        # Otherwise if the outermost bad annulus is NOT right at the end (either end) of the profile, we merge
        #  to the right
        else:
            cur_rads = np.delete(cur_rads, bad_cnts[-1])
            ann_masks = annular_mask(pix_centre, cur_rads[:-1], cur_rads[1:], rt.shape) * corr_mask[..., None]

        if ann_masks.shape[2] == 4 and not acceptable:
            warn("The requested annuli for {s} cannot be created, the data quality is too low. As such a set "
                 "of four annuli will be returned".format(s=source.name), stacklevel=2)
            break
    # Now of course, pixels must become a more useful unit again
    final_rads = (Quantity(cur_rads, 'pix') * pix_to_deg).to("arcsec")

    return final_rads, cnts, max_ann


def min_snr_proj_temp_prof(sources: Union[GalaxyCluster, ClusterSample], outer_radii: Union[Quantity, List[Quantity]],
                           min_snr: float = 20, min_width: Quantity = Quantity(20, 'arcsec'), use_combined: bool = True,
                           use_worst: bool = False, lo_en: Quantity = Quantity(0.5, 'keV'),
                           hi_en: Quantity = Quantity(2, 'keV'), psf_corr: bool = False, psf_model: str = "ELLBETA",
                           psf_bins: int = 4, psf_algo: str = "rl", psf_iter: int = 15, allow_negative: bool = False,
                           exp_corr: bool = True, group_spec: bool = True, min_counts: int = 5, min_sn: float = None,
                           over_sample: float = None, one_rmf: bool = True, freeze_met: bool = True,
                           abund_table: str = "angr", temp_lo_en: Quantity = Quantity(0.3, 'keV'),
                           temp_hi_en: Quantity = Quantity(7.9, 'keV'), num_cores: int = NUM_CORES) -> List[Quantity]:
    """
    This is a convenience function that allows you to quickly and easily start measuring projected
    temperature profiles of galaxy clusters, deciding on the annular bins using signal to noise measurements
    from photometric products. This function calls single_temp_apec_profile, but doesn't expose all of the more
    in depth variables, so if you want more control then use single_temp_apec_profile directly. The projected
    temperature profiles which are generated are added to their source's storage structure.

    :param GalaxyCluster/ClusterSample sources: An individual or sample of sources to measure projected
        temperature profiles for.
    :param str/Quantity outer_radii: The name or value of the outer radius to use for the generation of
        the spectra (for instance 'r200' would be acceptable for a GalaxyCluster, or Quantity(1000, 'kpc')). If
        'region' is chosen (to use the regions in region files), then any inner radius will be ignored. If you are
        generating for multiple sources then you can also pass a Quantity with one entry per source.
    :param float min_snr: The minimum signal to noise which is allowable in a given annulus.
    :param Quantity min_width: The minimum allowable width of an annulus. The default is set to 20 arcseconds to try
        and avoid PSF effects.
    :param bool use_combined: If True then the combined RateMap will be used for signal to noise annulus
        calculations, this is overridden by use_worst.
    :param bool use_worst: If True then the worst observation of the cluster (ranked by global signal to noise) will
        be used for signal to noise annulus calculations.
    :param Quantity lo_en: The lower energy bound of the ratemap to use for the signal to noise calculations.
    :param Quantity hi_en: The upper energy bound of the ratemap to use for the signal to noise calculations.
    :param bool psf_corr: Sets whether you wish to use a PSF corrected ratemap or not.
    :param str psf_model: If the ratemap you want to use is PSF corrected, this is the PSF model used.
    :param int psf_bins: If the ratemap you want to use is PSF corrected, this is the number of PSFs per
        side in the PSF grid.
    :param str psf_algo: If the ratemap you want to use is PSF corrected, this is the algorithm used.
    :param int psf_iter: If the ratemap you want to use is PSF corrected, this is the number of iterations.
    :param bool allow_negative: Should pixels in the background subtracted count map be allowed to go below
        zero, which results in a lower signal to noise (and can result in a negative signal to noise).
    :param bool exp_corr: Should signal to noises be measured with exposure time correction, default is True. I
            recommend that this be true for combined observations, as exposure time could change quite dramatically
            across the combined product.
    :param bool group_spec: A boolean flag that sets whether generated spectra are grouped or not.
    :param float min_counts: If generating a grouped spectrum, this is the minimum number of counts per channel.
        To disable minimum counts set this parameter to None.
    :param float min_sn: If generating a grouped spectrum, this is the minimum signal to noise in each channel.
        To disable minimum signal to noise set this parameter to None.
    :param float over_sample: The minimum energy resolution for each group, set to None to disable. e.g. if
        over_sample=3 then the minimum width of a group is 1/3 of the resolution FWHM at that energy.
    :param bool one_rmf: This flag tells the method whether it should only generate one RMF for a particular
        ObsID-instrument combination - this is much faster in some circumstances, however the RMF does depend
        slightly on position on the detector.
    :param bool freeze_met: Whether the metallicity parameter in the fits to annuli in XSPEC should be frozen.
    :param str abund_table: The abundance table to use during the XSPEC fits.
    :param Quantity temp_lo_en: The lower energy limit for the XSPEC fits to annular spectra.
    :param Quantity temp_hi_en: The upper energy limit for the XSPEC fits to annular spectra.
    :param int num_cores: The number of cores to use (if running locally), default is set to 90% of available.
    :return: A list of non-scalar astropy quantities containing the annular radii used to generate the
        projected temperature profiles created by this function. Each Quantity element of the list corresponds
        to a source.
    :rtype: List[Quantity]
    """

    if outer_radii != 'region':
        inn_rad_vals, out_rad_vals = region_setup(sources, outer_radii, Quantity(0, 'arcsec'), True, '')[1:]
    else:
        raise NotImplementedError("I don't currently support fitting region spectra")

    if all([use_combined, use_worst]):
        warn("You have passed both use_combined and use_worst as True. use_worst overrides use_combined, so the "
             "worst observation for each source will be used to decide on the annuli.")
        use_combined = False
    elif all([not use_combined, not use_worst]):
        warn("You have passed both use_combined and use_worst as False. One of them must be True, so here we default"
             " to using the combined data to decide on the annuli.")
        use_combined = True

    if abund_table not in ABUND_TABLES:
        avail_abund = ", ".join(ABUND_TABLES)
        raise ValueError("{a} is not a valid abundance table choice, please use one of the "
                         "following; {av}".format(a=abund_table, av=avail_abund))

    if isinstance(sources, BaseSource):
        sources = [sources]

    all_rads = []
    for src_ind, src in enumerate(sources):
        if use_combined:
            # This is the simplest option, we just use the combined ratemap to decide on the annuli with minimum SNR
            rads, snrs, ma = _snr_bins(src, out_rad_vals[src_ind], min_snr, min_width, lo_en, hi_en, psf_corr=psf_corr,
                                       psf_model=psf_model, psf_bins=psf_bins, psf_algo=psf_algo, psf_iter=psf_iter,
                                       allow_negative=allow_negative, exp_corr=exp_corr)
        else:
            # This way is slightly more complicated, but here we use the worst observation (ranked by global
            #  signal to noise).
            # The return for this function is ranked worst to best, so we grab the first row (which is an ObsID and
            #  instrument), then call _snr_bins with that one
            lowest_ranked = src.snr_ranking(out_rad_vals[src_ind], lo_en, hi_en, allow_negative)[0][0, :]
            rads, snrs, ma = _snr_bins(src, out_rad_vals[src_ind], min_snr, min_width, lo_en, hi_en, lowest_ranked[0],
                                       lowest_ranked[1], psf_corr, psf_model, psf_bins, psf_algo, psf_iter,
                                       allow_negative, exp_corr)

        # Shoves the annuli we've decided upon into a list for single_temp_apec_profile to use
        all_rads.append(rads)

    if len(sources) == 1:
        sources = sources[0]

    single_temp_apec_profile(sources, all_rads, group_spec=group_spec, min_counts=min_counts, min_sn=min_sn,
                             over_sample=over_sample, one_rmf=one_rmf, num_cores=num_cores, abund_table=abund_table,
                             lo_en=temp_lo_en, hi_en=temp_hi_en, freeze_met=freeze_met)

    return all_rads


def min_cnt_proj_temp_prof(sources: Union[GalaxyCluster, ClusterSample], outer_radii: Union[Quantity, List[Quantity]],
                           min_cnt: Union[int, Quantity] = Quantity(1000, 'ct'),
                           min_width: Quantity = Quantity(20, 'arcsec'), use_combined: bool = True,
                           lo_en: Quantity = Quantity(0.5, 'keV'), hi_en: Quantity = Quantity(2, 'keV'),
                           psf_corr: bool = False, psf_model: str = "ELLBETA", psf_bins: int = 4, psf_algo: str = "rl",
                           psf_iter: int = 15, group_spec: bool = True, min_counts: int = 5, min_sn: float = None,
                           over_sample: float = None, one_rmf: bool = True, freeze_met: bool = True,
                           abund_table: str = "angr", temp_lo_en: Quantity = Quantity(0.3, 'keV'),
                           temp_hi_en: Quantity = Quantity(7.9, 'keV'), num_cores: int = NUM_CORES) -> List[Quantity]:
    """
    This is a convenience function that allows you to quickly and easily start measuring projected
    temperature profiles of galaxy clusters, deciding on the annular bins using X-ray count measurements
    from photometric products. This function calls single_temp_apec_profile, but doesn't necessarily expose
    all of single_temp_apec_profile's variables, so if you want more control, then use single_temp_apec_profile
    directly. The projected temperature profiles which are generated are added to their source's storage structure.

    :param GalaxyCluster/ClusterSample sources: An individual or sample of sources to measure projected
        temperature profiles for.
    :param str/Quantity outer_radii: The name or value of the outer radius to use for the generation of
        the spectra (for instance 'r200' would be acceptable for a GalaxyCluster, or Quantity(1000, 'kpc')). If
        'region' is chosen (to use the regions in region files), then any inner radius will be ignored. If you are
        generating for multiple sources then you can also pass a Quantity with one entry per source.
    :param float min_cnt: The minimum counts allowable in a given annulus.
    :param Quantity min_width: The minimum allowable width of an annulus. The default is set to 20 arcseconds to try
        and avoid PSF effects.
    :param bool use_combined: If True then the combined RateMap will be used for count annulus calculations.
        Default is True, if False then the median ObsID-Instrument combo (in terms of background-subtracted counts
        within outer_radii) will be used to generate the annuli.
    :param Quantity lo_en: The lower energy bound of the ratemap to use for the count calculations.
    :param Quantity hi_en: The upper energy bound of the ratemap to use for the count calculations.
    :param bool psf_corr: Sets whether you wish to use a PSF corrected ratemap or not.
    :param str psf_model: If the ratemap you want to use is PSF corrected, this is the PSF model used.
    :param int psf_bins: If the ratemap you want to use is PSF corrected, this is the number of PSFs per
        side in the PSF grid.
    :param str psf_algo: If the ratemap you want to use is PSF corrected, this is the algorithm used.
    :param int psf_iter: If the ratemap you want to use is PSF corrected, this is the number of iterations.
    :param bool group_spec: A boolean flag that sets whether generated spectra are grouped or not.
    :param float min_counts: If generating a grouped spectrum, this is the minimum number of counts per channel.
        To disable minimum counts set this parameter to None.
    :param float min_sn: If generating a grouped spectrum, this is the minimum signal to noise in each channel.
        To disable minimum signal to noise set this parameter to None.
    :param float over_sample: The minimum energy resolution for each group, set to None to disable. e.g. if
        over_sample=3 then the minimum width of a group is 1/3 of the resolution FWHM at that energy.
    :param bool one_rmf: This flag tells the method whether it should only generate one RMF for a particular
        ObsID-instrument combination - this is much faster in some circumstances, however the RMF does depend
        slightly on position on the detector.
    :param bool freeze_met: Whether the metallicity parameter in the fits to annuli in XSPEC should be frozen.
    :param str abund_table: The abundance table to use during the XSPEC fits.
    :param Quantity temp_lo_en: The lower energy limit for the XSPEC fits to annular spectra.
    :param Quantity temp_hi_en: The upper energy limit for the XSPEC fits to annular spectra.
    :param int num_cores: The number of cores to use (if running locally), default is set to 90% of available.
    :return: A list of non-scalar astropy quantities containing the annular radii used to generate the
        projected temperature profiles created by this function. Each Quantity element of the list corresponds
        to a source.
    :rtype: List[Quantity]
    """

    if outer_radii != 'region':
        inn_rad_vals, out_rad_vals = region_setup(sources, outer_radii, Quantity(0, 'arcsec'), True, '')[1:]
    else:
        raise NotImplementedError("I don't currently support fitting region spectra")

    if abund_table not in ABUND_TABLES:
        avail_abund = ", ".join(ABUND_TABLES)
        raise ValueError("{a} is not a valid abundance table choice, please use one of the "
                         "following; {av}".format(a=abund_table, av=avail_abund))

    # Makes sure that sources is iterable, even if its just a single source - makes writing the rest of this
    #  function a bit neater.
    if isinstance(sources, BaseSource):
        sources = [sources]

    all_rads = []
    for src_ind, src in enumerate(sources):
        if use_combined:
            # This is the simplest option, we just use the combined ratemap to decide on the annuli with minimum counts
            rads, cnts, ma = _cnt_bins(src, out_rad_vals[src_ind], min_cnt, min_width, lo_en, hi_en, psf_corr=psf_corr,
                                       psf_model=psf_model, psf_bins=psf_bins, psf_algo=psf_algo, psf_iter=psf_iter)
        else:
            # Use the source's built in count ranking method (which in turn uses some RateMap class methods) to rank
            #  the individual observations (cnt_rnk is ObsID, Instrument combinations in order of ascending counts).
            # We then use the counts measured for each ObsID-Instrument combo (which are returned and stored in cnts)
            #  to decide upon the median observation.
            cnt_rnk, cnts = src.count_ranking(out_rad_vals[src_ind], lo_en, hi_en)
            # Obviously the median counts will not necessarily line up with any particular ObsID-instrument, but we
            #  can use the interpolation feature of numpy percentile to find the nearest existing counts to the
            #  median counts.
            med_obs_ind = np.argwhere(cnts == np.percentile(cnts, 50, interpolation='nearest'))[0]
            # This pulls out the ObsID and instrument that we have chosen to base the annular bins on.
            med_obs_id = cnt_rnk[med_obs_ind, 0][0]
            med_obs_inst = cnt_rnk[med_obs_ind, 1][0]

            # In this instance though, we use the median (in terms of background subtracted counts)
            #  individual (as in individual instrument too) observation to construct the annuli.
            rads, cnts, ma = _cnt_bins(src, out_rad_vals[src_ind], min_cnt, min_width, lo_en, hi_en, med_obs_id,
                                       med_obs_inst, psf_corr, psf_model, psf_bins, psf_algo, psf_iter)

        # Shoves the annuli we've decided upon into a list for single_temp_apec_profile to use
        all_rads.append(rads)

    # Reverses a bodge employed at the beginning of this function
    if len(sources) == 1:
        sources = sources[0]

    # This runs the fitting (and generation, if that has not already occurred) of the annular spectra.
    single_temp_apec_profile(sources, all_rads, group_spec=group_spec, min_counts=min_counts, min_sn=min_sn,
                             over_sample=over_sample, one_rmf=one_rmf, num_cores=num_cores, abund_table=abund_table,
                             lo_en=temp_lo_en, hi_en=temp_hi_en, freeze_met=freeze_met)

    return all_rads


def _grow_ann_proj_temp_prof(sources: Union[BaseSource, BaseSample], outer_radii: Union[Quantity, List[Quantity]],
                             growth_factor: float = 1.3, start_radius: Quantity = Quantity(20, 'arcsec'),
                             num_ann: int = None, group_spec: bool = True, min_counts: int = 5, min_sn: float = None,
                             over_sample: float = None, one_rmf: bool = True, num_cores: int = NUM_CORES):
    """
    This is a convenience function that allows you to quickly and easily start measuring projected temperature
    profiles of galaxy clusters where the outer radius of each annulus is some factor larger than that of the
    last annulus:
        .. math::
             R_{i+1} = R_{i}F

    If a growth factor is passed then the start radius and outer radius of a particular source will be used to solve
    for the number of annuli which should be generated. However if a number of annuli is passed (through num_ann),
    then this function will again use the start and outer radii and solve for the growth factor instead, over-riding
    any growth factor that may have been passed in.

    This function calls single_temp_apec_profile, but doesn't expose all of the more
    in depth variables, so if you want more control then use single_temp_apec_profile directly. The projected
    temperature profiles which are generated are added to their source's storage structure.

    :param GalaxyCluster/ClusterSample sources: An individual or sample of sources to measure projected
        temperature profiles for.
    :param str/Quantity outer_radii: The name or value of the outer radius to use for the generation of
        the spectra (for instance 'r200' would be acceptable for a GalaxyCluster, or Quantity(1000, 'kpc')). If
        'region' is chosen (to use the regions in region files), then any inner radius will be ignored. If you are
        generating for multiple sources then you can also pass a Quantity with one entry per source.
    :param float growth_factor: The factor by which the outer radius of the Nth annulus should be larger than
        the outer radius of the N-1th annulus. This will be over-ridden by a re-calculated value if a value
        is passed to num_ann.
    :param Quantity start_radius: The radius of the innermost circular annulus, the default is 20 arcseconds, which
        was chosen to try and avoid PSF effects.
    :param int num_ann: The number of annuli which should be used, default is None, in which case the value will be
        calculated using the growth factor, outer radius, and start radius. If this parameter is passed then
        growth_factor will be overridden by a recalculated value.
    :param bool group_spec: A boolean flag that sets whether generated spectra are grouped or not.
    :param float min_counts: If generating a grouped spectrum, this is the minimum number of counts per channel.
        To disable minimum counts set this parameter to None.
    :param float min_sn: If generating a grouped spectrum, this is the minimum signal to noise in each channel.
        To disable minimum signal to noise set this parameter to None.
    :param float over_sample: The minimum energy resolution for each group, set to None to disable. e.g. if
        over_sample=3 then the minimum width of a group is 1/3 of the resolution FWHM at that energy.
    :param bool one_rmf: This flag tells the method whether it should only generate one RMF for a particular
        ObsID-instrument combination - this is much faster in some circumstances, however the RMF does depend
        slightly on position on the detector.
    :param int num_cores: The number of cores to use (if running locally), default is set to 90% of available.
    """

    raise NotImplementedError("This doesn't work yet because I got bored")

    if outer_radii != 'region':
        inn_rad_vals, out_rad_vals = region_setup(sources, outer_radii, Quantity(0, 'arcsec'), True, '')[1:]
    else:
        raise NotImplementedError("I don't currently support fitting region spectra")

    all_rads = []
    for src_ind, src in enumerate(sources):
        cur_start = src.convert_radius(start_radius, 'arcsec')
        if num_ann is None:
            cur_num_ann = int(np.ceil(np.log(out_rad_vals[src_ind].to('arcsec').value / cur_start.value)
                                      / np.log(growth_factor)))
            cur_growth_factor = growth_factor
        else:
            cur_growth_factor = np.power(out_rad_vals[src_ind].to('arcsec').value / cur_start.value, 1 / num_ann)
            cur_num_ann = num_ann

        rads = [cur_start.value]
        rads += [cur_start.value * ann_ind * cur_growth_factor for ann_ind in range(1, cur_num_ann + 1)]
        print(Quantity(rads, 'arcsec'))
        print('')

    single_temp_apec_profile(sources, all_rads, group_spec=group_spec, min_counts=min_counts, min_sn=min_sn,
                             over_sample=over_sample, one_rmf=one_rmf, num_cores=num_cores)


def onion_deproj_temp_prof(sources: Union[GalaxyCluster, ClusterSample], outer_radii: Union[Quantity, List[Quantity]],
                           annulus_method: str = 'min_snr', min_snr: float = 30,
                           min_cnt: Union[int, Quantity] = Quantity(1000, 'ct'),
                           min_width: Quantity = Quantity(20, 'arcsec'), use_combined: bool = True,
                           use_worst: bool = False, lo_en: Quantity = Quantity(0.5, 'keV'),
                           hi_en: Quantity = Quantity(2, 'keV'), psf_corr: bool = False, psf_model: str = "ELLBETA",
                           psf_bins: int = 4, psf_algo: str = "rl", psf_iter: int = 15, allow_negative: bool = False,
                           exp_corr: bool = True, group_spec: bool = True, min_counts: int = 5, min_sn: float = None,
                           over_sample: float = None, one_rmf: bool = True, freeze_met: bool = True,
                           abund_table: str = "angr", temp_lo_en: Quantity = Quantity(0.3, 'keV'),
                           temp_hi_en: Quantity = Quantity(7.9, 'keV'), num_data_real: int = 3000,
                           conf_level: int = 68.2, num_cores: int = NUM_CORES) -> List[GasTemperature3D]:
    """
    This function will generate de-projected, three-dimensional, gas temperature profiles of galaxy clusters using
    the 'onion peeling' deprojection method. It will also generate any projected temperature profiles that may be
    necessary, using the method specified in the function call (the default is the minimum signal to noise annuli
    method). As a side effect of that process APEC normalisation profiles will also be created, as well as Emission
    Measure profiles. The function is an implementation of a fairly old technique, though it has been used recently
    in https://doi.org/10.1051/0004-6361/201731748. For a more in depth discussion of this technique and its uses
    I would currently recommend https://doi.org/10.1051/0004-6361:20020905.

    :param GalaxyCluster/ClusterSample sources: An individual or sample of sources to calculate 3D temperature
        profiles for.
    :param str/Quantity outer_radii: The name or value of the outer radius to use for the generation of
        the spectra (for instance 'r200' would be acceptable for a GalaxyCluster, or Quantity(1000, 'kpc')). If
        'region' is chosen (to use the regions in region files), then any inner radius will be ignored. If you are
        generating for multiple sources then you can also pass a Quantity with one entry per source.
    :param str annulus_method: The method by which the annuli are designated, this can be 'min_snr' (which will use
        the min_snr_proj_temp_prof function), or 'min_cnt' (which will use the min_cnt_proj_temp_prof function).
    :param float min_snr: The minimum signal-to-noise which is allowable in a given annulus, used if annulus_method
        is set to 'min_snr'.
    :param int/Quantity min_cnt: The minimum background subtracted counts which are allowable in a given annulus, used
        if annulus_method is set to 'min_cnt'.
    :param Quantity min_width: The minimum allowable width of an annulus. The default is set to 20 arcseconds to try
        and avoid PSF effects.
    :param bool use_combined: If True (and annulus_method is set to 'min_snr') then the combined RateMap will be
        used for signal-to-noise annulus calculations, this is overridden by use_worst. If True (and annulus_method
        is set to 'min_cnt') then combined RateMaps will be used for annulus count calculations, if False then
        the median observation (in terms of counts) will be used.
    :param bool use_worst: If True then the worst observation of the cluster (ranked by global signal-to-noise) will
        be used for signal-to-noise annulus calculations. Used if annulus_method is set to 'min_snr'.
    :param Quantity lo_en: The lower energy bound of the RateMap to use for the signal-to-noise or background
        subtracted count calculations.
    :param Quantity hi_en: The upper energy bound of the RateMap to use for the signal-to-noise or background
        subtracted count calculations.
    :param bool psf_corr: Sets whether you wish to use a PSF corrected RateMap or not.
    :param str psf_model: If the RateMap you want to use is PSF corrected, this is the PSF model used.
    :param int psf_bins: If the RateMap you want to use is PSF corrected, this is the number of PSFs per
        side in the PSF grid.
    :param str psf_algo: If the RateMap you want to use is PSF corrected, this is the algorithm used.
    :param int psf_iter: If the RateMap you want to use is PSF corrected, this is the number of iterations.
    :param bool allow_negative: Should pixels in the background subtracted count map be allowed to go below
        zero, which results in a lower signal-to-noise (and can result in a negative signal-to-noise).
    :param bool exp_corr: Should signal to noises be measured with exposure time correction, default is True. I
            recommend that this be true for combined observations, as exposure time could change quite dramatically
            across the combined product.
    :param bool group_spec: A boolean flag that sets whether generated spectra are grouped or not.
    :param float min_counts: If generating a grouped spectrum, this is the minimum number of counts per channel.
        To disable minimum counts set this parameter to None.
    :param float min_sn: If generating a grouped spectrum, this is the minimum signal to noise in each channel.
        To disable minimum signal to noise set this parameter to None.
    :param float over_sample: The minimum energy resolution for each group, set to None to disable. e.g. if
        over_sample=3 then the minimum width of a group is 1/3 of the resolution FWHM at that energy.
    :param bool one_rmf: This flag tells the method whether it should only generate one RMF for a particular
        ObsID-instrument combination - this is much faster in some circumstances, however the RMF does depend
        slightly on position on the detector.
    :param bool freeze_met: Whether the metallicity parameter in the fits to annuli in XSPEC should be frozen.
    :param str abund_table: The abundance table to use both for the conversion from n_exn_p to n_e^2 during density
        calculation, and the XSPEC fit.
    :param Quantity temp_lo_en: The lower energy limit for the XSPEC fits to annular spectra.
    :param Quantity temp_hi_en: The upper energy limit for the XSPEC fits to annular spectra.
    :param int num_data_real: The number of random realisations to generate when propagating profile
        uncertainties, the default is 3000.
    :param int conf_level: What sigma uncertainties should newly created profiles have, the default is 1σ.
    :param int num_cores: The number of cores to use (if running locally), default is set to 90% of available.
    :return: A list of the 3D temperature profiles measured by this function, though if the measurement was not
        successful an entry of None will be added to the list.
    :rtype: List[GasTemperature3D]
    """
    if annulus_method not in ALLOWED_ANN_METHODS:
        a_meth = ", ".join(ALLOWED_ANN_METHODS)
        raise ValueError("That is not a valid method for deciding where to place annuli, please use one of "
                         "these; {}".format(a_meth))

    if annulus_method == 'min_snr':
        # This returns the boundary radii for the annuli
        ann_rads = min_snr_proj_temp_prof(sources, outer_radii, min_snr, min_width, use_combined, use_worst, lo_en,
                                          hi_en, psf_corr, psf_model, psf_bins, psf_algo, psf_iter, allow_negative,
                                          exp_corr, group_spec, min_counts, min_sn, over_sample, one_rmf, freeze_met,
                                          abund_table, temp_lo_en, temp_hi_en, num_cores)
    elif annulus_method == 'min_cnt':
        # This returns the boundary radii for the annuli, based on a minimum number of counts per annulus
        ann_rads = min_cnt_proj_temp_prof(sources, outer_radii, min_cnt, min_width, use_combined, lo_en, hi_en,
                                          psf_corr, psf_model, psf_bins, psf_algo, psf_iter, group_spec, min_counts,
                                          min_sn, over_sample, one_rmf, freeze_met, abund_table, temp_lo_en, temp_hi_en,
                                          num_cores)
    elif annulus_method == "growth":
        raise NotImplementedError("This method isn't implemented yet")

    # So we can iterate through sources without worrying if there's more than one cluster
    if not isinstance(sources, (BaseSample, list)):
        sources = [sources]

    all_3d_temp_profs = []
    # Don't need to check abundance table input because that happens in min_snr_proj_temp_prof
    for src_ind, src in enumerate(sources):
        cur_rads = ann_rads[src_ind]

        try:
            # The projected temperature profile we're going to use
            proj_temp = src.get_proj_temp_profiles(cur_rads, group_spec, min_counts, min_sn, over_sample)
            # The normalisation profile(s) from the fit that produced the projected temperature profile.
            apec_norm_prof = src.get_apec_norm_profiles(cur_rads, group_spec, min_counts, min_sn, over_sample)
        except NoProductAvailableError:
            warn("{s} doesn't have a matching projected temperature profile, skipping.")
            all_3d_temp_profs.append(None)
            continue

        # We need to check if a matching 3D temperature profile has already been generated, as then we
        #  just use that one rather than making another (which would be silly and also eat up more storage
        #  because they are automatically saved to disk).
        try:
            existing_3d_temp_prof = src.get_3d_temp_profiles(set_id=proj_temp.set_ident)
            all_3d_temp_profs.append(existing_3d_temp_prof)
            continue
        except NoProductAvailableError:
            pass

        obs_id = 'combined'
        inst = 'combined'
        # There are reasons that a projected temperature profile can be considered unusable, so we must check. Also
        #  make sure to only use those profiles that have a minimum of 4 annuli. The len operator retrieves the number
        #  of radial data points a profile has
        if proj_temp.usable and len(proj_temp) > 3:
            # Also make an Emission Measure profile, used for weighting the contributions from different
            #  shells to annuli
            em_prof = apec_norm_prof.emission_measure_profile(src.redshift, src.cosmo, abund_table,
                                                              num_data_real, conf_level)
            src.update_products(em_prof)

            # Need to make sure the annular boundaries are a) in a proper distance unit rather than degrees, and b)
            #  in units of centimeters
            cur_rads = src.convert_radius(cur_rads, 'cm')
            # Use a handy function I wrote to calculate the volume intersections of spherical shells and
            #  projected annuli
            vol_intersects = shell_ann_vol_intersect(cur_rads, cur_rads)

            # Then it's an inverse matrix problem to recover the 3D temperatures
            temp_3d = (np.linalg.inv(vol_intersects.T) @ (proj_temp.values * em_prof.values)) / (np.linalg.inv(
                vol_intersects.T) @ em_prof.values)

            # I generate random realisations of the projected temperature profile and the emission measure profile
            #  to help me with error propagation
            proj_temp_reals = proj_temp.generate_data_realisations(num_data_real, truncate_zero=True)
            em_reals = em_prof.generate_data_realisations(num_data_real, truncate_zero=True)

            # Set up an N x R array for the random realisations of the 3D temperature, where N is the number
            #  of realisations and R is the number of radius data points
            temp_3d_reals = Quantity(np.zeros(proj_temp_reals.shape), proj_temp_reals.unit)
            for i in range(0, num_data_real):
                # Calculate and store the 3D temperature profile realisations
                interim = (np.linalg.inv(vol_intersects.T) @ (proj_temp_reals[i, :] * em_reals[i, :])) / (np.linalg.inv(
                    vol_intersects.T) @ em_reals[i, :])
                temp_3d_reals[i, :] = interim

            # Calculate the upper and lower confidence level values as specified by the user
            med = np.percentile(temp_3d_reals, 50, axis=0)
            upper = np.percentile(temp_3d_reals, 50 + (conf_level / 2), axis=0)
            lower = np.percentile(temp_3d_reals, 50 - (conf_level / 2), axis=0)

            # Bit dodgy to do this but oh well
            temp_3d_sigma = Quantity(np.average([med - lower, upper - med], axis=0), 'keV')

            # And finally actually set up a 3D temperature profile
            temp_3d_prof = GasTemperature3D(proj_temp.radii, temp_3d, proj_temp.centre, src.name, obs_id, inst,
                                            proj_temp.radii_err, temp_3d_sigma, proj_temp.set_ident,
                                            proj_temp.associated_set_storage_key, proj_temp.deg_radii,
                                            auto_save=True)
            src.update_products(temp_3d_prof)
            all_3d_temp_profs.append(temp_3d_prof)

        else:
            warn("The projected temperature profile for {src} is not considered usable by XGA".format(src=src.name),
                 stacklevel=2)
            all_3d_temp_profs.append(None)

    return all_3d_temp_profs



#  This code is a part of X-ray: Generate and Analyse (XGA), a module designed for the XMM Cluster Survey (XCS).
#  Last modified by David J Turner (turne540@msu.edu) 11/10/2024, 16:42. Copyright (c) The Contributors

from typing import Union, List, Tuple, Dict
from warnings import warn, simplefilter

import numpy as np
from astropy import wcs
from astropy.cosmology import Cosmology
from astropy.units import Quantity, UnitConversionError, kpc

from .general import ExtendedSource
from .. import DEFAULT_COSMO
from ..exceptions import NoRegionsError, NoProductAvailableError
from ..imagetools import radial_brightness
from ..products import Spectrum, BaseProfile1D
from ..products.profile import ProjectedGasTemperature1D, APECNormalisation1D, GasDensity3D, GasTemperature3D, \
    HydrostaticMass
from ..sourcetools import ang_to_rad, rad_to_ang

# This disables an annoying astropy warning that pops up all the time with XMM images
# Don't know if I should do this really
simplefilter('ignore', wcs.FITSFixedWarning)


class GalaxyCluster(ExtendedSource):
    """
    This class is for the declaration and analysis of GalaxyCluster sources, and is a subclass of ExtendedSource.
    Using this source for cluster analysis gives you access to many more analyses than an ExtendedSource, and due
    to the passing of overdensity radii it has a more physically based idea of the size of the object. There are
    also extra source matching steps to deal with the possible detection of cool cores as point sources, which can
    be turned on and off.

    :param float ra: The right-ascension of the cluster, in degrees.
    :param float dec: The declination of the cluster, in degrees.
    :param float redshift: The redshift of the cluster, required for cluster analysis.
    :param str name: The name of the cluster, optional. Name will be constructed from position if None.
    :param Quantity r200: A value for the R200 of the source. At least one overdensity radius must be passed.
    :param Quantity r500: A value for the R500 of the source. At least one overdensity radius must be passed.
    :param Quantity r2500: A value for the R2500 of the source. At least one overdensity radius must be passed.
    :param richness: An optical richness of the cluster, optional.
    :param richness_err: An uncertainty on the optical richness of the cluster, optional.
    :param Quantity wl_mass: A weak lensing mass of the cluster, optional.
    :param Quantity wl_mass_err: An uncertainty on the weak lensing mass of the cluster, optional.
    :param Quantity custom_region_radius: A custom analysis region radius for this cluster, optional.
    :param bool use_peak: Whether peak position should be found and used.
    :param Quantity peak_lo_en: The lower energy bound for the RateMap to calculate peak position
        from. Default is 0.5keV
    :param Quantity peak_hi_en: The upper energy bound for the RateMap to calculate peak position
        from. Default is 2.0keV.
    :param float back_inn_rad_factor: This factor is multiplied by an analysis region radius, and gives the inner
        radius for the background region. Default is 1.05.
    :param float back_out_rad_factor: This factor is multiplied by an analysis region radius, and gives the outer
        radius for the background region. Default is 1.5.
    :param cosmology: An astropy cosmology object for use throughout analysis of the source.
    :param bool load_products: Whether existing products should be loaded from disk.
    :param bool load_fits: Whether existing fits should be loaded from disk.
    :param str peak_find_method: Which peak finding method should be used (if use_peak is True). Default
        is hierarchical, simple may also be passed.
    :param bool clean_obs: Should the observations be subjected to a minimum coverage check, i.e. whether a
        certain fraction of a certain region is covered by an ObsID. Default is True.
    :param str clean_obs_reg: The region to use for the cleaning step, default is R200.
    :param float clean_obs_threshold: The minimum coverage fraction for an observation to be kept for analysis.
    :param bool regen_merged: Should merged images/exposure maps be regenerated after cleaning. Default is True.
    :param str peak_find_method: Which peak finding method should be used (if use_peak is True). Default
        is hierarchical, simple may also be passed.
    :param bool in_sample: A boolean argument that tells the source whether it is part of a sample or not, setting
        to True suppresses some warnings so that they can be displayed at the end of the sample progress bar. Default
        is False. User should only set to True to remove warnings.
    :param bool include_core_pnt_srcs: Controls whether bright point sources near the user-defined coordinate
        (hopefully the core) are allowed to contribute to analyses. Default is True, in which case such point
        sources will NOT be included in masks, in case they are a bright cool core.
    """
    def __init__(self, ra, dec, redshift, name=None, r200: Quantity = None, r500: Quantity = None,
                 r2500: Quantity = None, richness: float = None, richness_err: float = None,
                 wl_mass: Quantity = None, wl_mass_err: Quantity = None, custom_region_radius=None, use_peak=True,
                 peak_lo_en=Quantity(0.5, "keV"), peak_hi_en=Quantity(2.0, "keV"), back_inn_rad_factor=1.05,
                 back_out_rad_factor=1.5, cosmology: Cosmology = DEFAULT_COSMO, load_products=True, load_fits=False,
                 clean_obs=True, clean_obs_reg="r200", clean_obs_threshold=0.3, regen_merged: bool = True,
                 peak_find_method: str = "hierarchical", in_sample: bool = False, include_core_pnt_srcs: bool = True):
        """
        The init of the GalaxyCluster specific XGA class, takes information on the cluster to enable analyses.

        :param float ra: The right-ascension of the cluster, in degrees.
        :param float dec: The declination of the cluster, in degrees.
        :param float redshift: The redshift of the cluster, required for cluster analysis.
        :param str name: The name of the cluster, optional. Name will be constructed from position if None.
        :param Quantity r200: A value for the R200 of the source. At least one overdensity radius must be passed.
        :param Quantity r500: A value for the R500 of the source. At least one overdensity radius must be passed.
        :param Quantity r2500: A value for the R2500 of the source. At least one overdensity radius must be passed.
        :param richness: An optical richness of the cluster, optional.
        :param richness_err: An uncertainty on the optical richness of the cluster, optional.
        :param Quantity wl_mass: A weak lensing mass of the cluster, optional.
        :param Quantity wl_mass_err: An uncertainty on the weak lensing mass of the cluster, optional.
        :param Quantity custom_region_radius: A custom analysis region radius for this cluster, optional.
        :param bool use_peak: Whether peak position should be found and used.
        :param Quantity peak_lo_en: The lower energy bound for the RateMap to calculate peak position
            from. Default is 0.5keV
        :param Quantity peak_hi_en: The upper energy bound for the RateMap to calculate peak position
            from. Default is 2.0keV.
        :param float back_inn_rad_factor: This factor is multiplied by an analysis region radius, and gives the inner
            radius for the background region. Default is 1.05.
        :param float back_out_rad_factor: This factor is multiplied by an analysis region radius, and gives the outer
            radius for the background region. Default is 1.5.
        :param cosmology: An astropy cosmology object for use throughout analysis of the source.
        :param bool load_products: Whether existing products should be loaded from disk.
        :param bool load_fits: Whether existing fits should be loaded from disk.
        :param str peak_find_method: Which peak finding method should be used (if use_peak is True). Default
            is hierarchical, simple may also be passed.
        :param bool clean_obs: Should the observations be subjected to a minimum coverage check, i.e. whether a
            certain fraction of a certain region is covered by an ObsID. Default is True.
        :param str clean_obs_reg: The region to use for the cleaning step, default is R200.
        :param float clean_obs_threshold: The minimum coverage fraction for an observation to be kept for analysis.
        :param bool regen_merged: Should merged images/exposure maps be regenerated after cleaning. Default is True.
        :param str peak_find_method: Which peak finding method should be used (if use_peak is True). Default
            is hierarchical, simple may also be passed.
        :param bool in_sample: A boolean argument that tells the source whether it is part of a sample or not, setting
            to True suppresses some warnings so that they can be displayed at the end of the sample progress
            bar. Default is False. User should only set to True to remove warnings.
        :param bool include_core_pnt_srcs: Controls whether bright point sources near the user-defined coordinate
            (hopefully the core) are allowed to contribute to analyses. Default is True, in which case such point
            sources will NOT be included in masks, in case they are a bright cool core.

        """
        # Store the passed value of 'include_core_pnt_srcs' in an attribute now, before we run the super-class init,
        #  as we want the GalaxyCluster _source_type_match method to be able to use it to determine if point
        #  sources that might be a bright cool-core are included or not
        self._include_core_pnt = include_core_pnt_srcs

        # Set up the radii dictionary, also before the superclass init call.
        self._radii = {}
        if r200 is None and r500 is None and r2500 is None:
            raise ValueError("You must set at least one overdensity radius")

        # Here we don't need to check if a non-null redshift was supplied, a redshift is required for
        #  initialising a GalaxyCluster object. These chunks just convert the radii to kpc.
        # I know its ugly to have the same code three times, but I want these to be in attributes.
        if r200 is not None and r200.unit.is_equivalent("deg"):
            self._r200 = ang_to_rad(r200, redshift, cosmology).to("kpc")
            # Radii must be stored in degrees in the internal radii dictionary
            self._radii["r200"] = r200.to("deg")
        elif r200 is not None and r200.unit.is_equivalent("kpc"):
            self._r200 = r200.to("kpc")
            self._radii["r200"] = rad_to_ang(r200, redshift, cosmology)
        elif r200 is not None and not r200.unit.is_equivalent("kpc") and not r200.unit.is_equivalent("deg"):
            raise UnitConversionError("R200 radius must be in either angular or distance units.")
        elif r200 is None and clean_obs_reg == "r200":
            clean_obs_reg = "r500"

        if r500 is not None and r500.unit.is_equivalent("deg"):
            self._r500 = ang_to_rad(r500, redshift, cosmology).to("kpc")
            self._radii["r500"] = r500.to("deg")
        elif r500 is not None and r500.unit.is_equivalent("kpc"):
            self._r500 = r500.to("kpc")
            self._radii["r500"] = rad_to_ang(r500, redshift, cosmology)
        elif r500 is not None and not r500.unit.is_equivalent("kpc") and not r500.unit.is_equivalent("deg"):
            raise UnitConversionError("R500 radius must be in either angular or distance units.")

        if r2500 is not None and r2500.unit.is_equivalent("deg"):
            self._r2500 = ang_to_rad(r2500, redshift, cosmology).to("kpc")
            self._radii["r2500"] = r2500.to("deg")
        elif r2500 is not None and r2500.unit.is_equivalent("kpc"):
            self._r2500 = r2500.to("kpc")
            self._radii["r2500"] = rad_to_ang(r2500, redshift, cosmology)
        elif r2500 is not None and not r2500.unit.is_equivalent("kpc") and not r2500.unit.is_equivalent("deg"):
            raise UnitConversionError("R2500 radius must be in either angular or distance units.")

        super().__init__(ra, dec, redshift, name, custom_region_radius, use_peak, peak_lo_en, peak_hi_en,
                         back_inn_rad_factor, back_out_rad_factor, cosmology, load_products, load_fits,
                         peak_find_method, in_sample)

        # Reading observables into their attributes, if the user doesn't pass a value for a particular observable
        #  it will be None.
        self._richness = richness
        self._richness_err = richness_err

        # Mass has a unit, unlike richness, so need to check that as we're reading it in
        if wl_mass is not None and wl_mass.unit.is_equivalent("Msun"):
            self._wl_mass = wl_mass.to("Msun")
        elif wl_mass is not None and not wl_mass.unit.is_equivalent("Msun"):
            raise UnitConversionError("The weak lensing mass value cannot be converted to MSun.")

        if wl_mass_err is not None and wl_mass_err.unit.is_equivalent("Msun"):
            self._wl_mass_err = wl_mass_err.to("Msun")
        elif wl_mass_err is not None and not wl_mass_err.unit.is_equivalent("Msun"):
            raise UnitConversionError("The weak lensing mass error value cannot be converted to MSun.")

        if clean_obs and clean_obs_reg in self._radii:
            # Use this method to figure out what data to throw away
            reject_dict = self.obs_check(clean_obs_reg, clean_obs_threshold)
            if len(reject_dict) != 0:
                # Use the source method to remove data we've decided isn't worth keeping
                self.disassociate_obs(reject_dict)
                # I used to run these just so there is an up to date combined ratemap, but its quite
                #  inefficient to do it on an individual basis if dealing with a sample, so the user will have
                #  to run those commands themselves later
                # Now I will run them only if the regen_merged flag is True
                if regen_merged:
                    from ..sas import emosaic
                    emosaic(self, "image", self._peak_lo_en, self._peak_hi_en, disable_progress=True)
                    emosaic(self, "expmap", self._peak_lo_en, self._peak_hi_en, disable_progress=True)
                    self._all_peaks(peak_find_method)

                    # And finally this sets the default coordinate to the peak if use peak is True
                    if self._use_peak:
                        self._default_coord = self.peak

        # Throws an error if a poor choice of region has been made
        elif clean_obs and clean_obs_reg not in self._radii:
            raise NoRegionsError("{c} is not associated with {s}".format(c=clean_obs_reg, s=self.name))

    def _source_type_match(self, source_type: str) -> Tuple[Dict, Dict, Dict]:
        """
        A function to override the _source_type_match method of the BaseSource class, containing slightly
        more complex matching criteria for galaxy clusters. Galaxy clusters having their own version of this
        method was driven by issue #407, the problems I was having with low redshift clusters particularly.

        Point sources within 0.15R500, 0.15*0.66*R200, or 0.15*2.25*R2500 (in order of descending priority, R200
        will only be used if R500 isn't available etc. - the extra factors for R200 and R2500 are meant to convert
        the radius to ~R500, and were arrived at using the Arnaud et al. 2005 R-T scaling relations) will be allowed
        to remain in the analysis, as they may well be cool-cores.

        This method also attempts to check for fragmentation of clusters by the source finder, which can cause
        issues where low redshift clusters are split up into multiple extended sources. Any interloper sources which
        are extended and intersect with a source region that has been designated the actual source will NOT be
        removed from the analysis - in practise this may be more use in making sure regions which are not consistent
        across observations do not remove chunks of the cluster (see issue #407).

        :param str source_type: Should either be ext or pnt, describes what type of source I
            should be looking for in the region files.
        :return: A dictionary containing the matched region for each ObsID + a combined region, another
            dictionary containing any sources that matched to the coordinates and weren't chosen,
            and a final dictionary with sources that aren't the target, or in the 2nd dictionary.
        :rtype: Tuple[Dict, Dict, Dict]
        """
        def dist_from_source(reg):
            """
            Calculates the euclidean distance between the centre of a supplied region, and the
            position of the source.

            :param reg: A region object.
            :return: Distance between region centre and source position.
            """
            ra = reg.center.ra.value
            dec = reg.center.dec.value
            return Quantity(np.sqrt(abs(ra - self._ra_dec[0]) ** 2 + abs(dec - self._ra_dec[1]) ** 2), 'deg')

        results_dict, alt_match_dict, anti_results_dict = super()._source_type_match('ext')

        # The 0.66 and 2.25 factors are intended to shift the r200 and r2500 values to approximately r500, and were
        #  decided on by dividing the Arnaud et al. 2005 R-T relations by one another and finding the mean factor
        if 'r500' in self._radii:
            check_rad = self.convert_radius(self._radii['r500'] * 0.15, 'deg')
        elif 'r200' in self._radii:
            check_rad = self.convert_radius(self._radii['r200'] * 0.66 * 0.15, 'deg')
        else:
            check_rad = self.convert_radius(self._radii['r2500'] * 2.25 * 0.15, 'deg')

        # Here we scrub the anti-results dictionary (I don't know why I called it that...) to make sure cool cores
        #  aren't accidentally removed, and that chunks of cluster emission aren't removed.
        # However, the initialization of a GalaxyCluster lets the user control whether point sources near the core
        #  should be left in (the default behaviour) or still removed. This is a tiny little refinement I've been
        #  meaning to add for literally years...
        new_anti_results = {}
        for obs in self._obs:
            # This is where the cleaned interlopers will be stored
            new_anti_results[obs] = []
            # Cycling through the current interloper regions for the current ObsID
            for reg_obj in anti_results_dict[obs]:
                # Calculating the distance (in degrees) of the centre of the current interloper region from
                #  the user supplied coordinates of the cluster
                dist = dist_from_source(reg_obj)

                # If the current interloper source is a point source/a PSF sized extended source and is within the
                #  fraction of the chosen characteristic radius of the cluster then we assume it is a poorly handled
                #  cool core and allow it to stay in the analysis - OR THE USER CAN JUST HAVE TURNED THIS BEHAVIOUR
                #  OFF WITH include_core_pnt
                if self._include_core_pnt and reg_obj.visual["edgecolor"] == 'red' and dist < check_rad:
                    warn_text = "A point source has been detected in {o} and is very close to the user supplied " \
                                "coordinates of {s}. It will not be excluded from analysis due to the possibility " \
                                "of a mis-identified cool core".format(s=self.name, o=obs)
                    if not self._samp_member:
                        # We do print a warning though
                        warn(warn_text, stacklevel=2)
                    else:
                        self._supp_warn.append(warn_text)

                elif self._include_core_pnt and reg_obj.visual["edgecolor"] == "magenta" and dist < check_rad:
                    warn_text = "A PSF sized extended source has been detected in {o} and is very close to the " \
                                "user supplied coordinates of {s}. It will not be excluded from analysis due " \
                                "to the possibility of a mis-identified cool core".format(s=self.name, o=obs)
                    if not self._samp_member:
                        warn(warn_text, stacklevel=2)
                    else:
                        self._supp_warn.append(warn_text)
                else:
                    new_anti_results[obs].append(reg_obj)

            # Here we run through the 'chosen' region for each observation (so the region that we think is the
            #  cluster) and check if any of the current set of interloper regions intersects with it. If they do
            #  then they are allowed to stay in the analysis under assumption that they're actually part of the
            #  cluster
            for res_obs in results_dict:
                if results_dict[res_obs] is not None:
                    # Reads out the chosen region for res_obs
                    src_reg_obj = results_dict[res_obs]
                    # Stores its central coordinates in an astropy quantity
                    centre = Quantity([src_reg_obj.center.ra.value, src_reg_obj.center.dec.value], 'deg')

                    # At first I set the checking radius to the semimajor axis
                    rad = Quantity(src_reg_obj.width.to('deg').value/2, 'deg')
                    # And use my handy method to find which regions intersect with a circle with the semimajor length
                    #  as radius, centred on the centre of the current chosen region
                    within_width = self.regions_within_radii(Quantity(0, 'deg'), rad, centre, new_anti_results[obs])
                    # Make sure to only select extended (green) sources
                    within_width = [reg for reg in within_width if reg.visual['edgecolor'] == 'green']

                    # Then I repeat that process with the semiminor axis, and if a interloper intersects with both
                    #  then it would intersect with the ellipse of the current chosen region.
                    rad = Quantity(src_reg_obj.height.to('deg').value/2, 'deg')
                    within_height = self.regions_within_radii(Quantity(0, 'deg'), rad, centre, new_anti_results[obs])
                    within_height = [reg for reg in within_height if reg.visual['edgecolor'] == 'green']
                    
                    # This finds which regions are present in both lists and makes sure if they are in both
                    #  then they are NOT removed from the analysis - AS OF regions v0.9 THIS NO LONGER WORKS AS
                    #  REGIONS ARE NOT HASHABLE - THE LIST COMPREHENSION BELOW IS A QUICK FIX BUT LESS EFFICIENT
                    # intersect_regions = list(set(within_width) & set(within_height))
                    
                    # This should do what the above set intersection did, but slower
                    intersect_regions = [r for r in within_height if r in within_width]
                    for inter_reg in intersect_regions:
                        inter_reg_ind = new_anti_results[obs].index(inter_reg)
                        new_anti_results[obs].pop(inter_reg_ind)

        return results_dict, alt_match_dict, new_anti_results

    def _new_rad_checks(self, new_rad: Quantity) -> Tuple[Quantity, Quantity]:
        """
        An internal method to check and convert new values of overdensity radii passed to the setters
        of those overdensity radii properties. Purely to avoid repeating the same code multiple times.

        :param Quantity new_rad: The new radius that the user has passed to the property setter.
        :return: The radius converted to kpc, and to degrees.
        :rtype: Tuple[Quantity, Quantity]
        """
        if not isinstance(new_rad, Quantity):
            raise TypeError("New overdensity radii must be an astropy Quantity")

        # This will make sure that the radius is converted into kpc, and will throw errors if the new_rad is in
        #  stupid units, so I don't need to do those checks here.
        converted_kpc_rad = self.convert_radius(new_rad, 'kpc')
        # For some reason I setup the _radii internal dictionary to have units of degrees, so I convert to that as well
        converted_deg_rad = self.convert_radius(new_rad, 'deg')

        return converted_kpc_rad, converted_deg_rad

    # Property getters for the over density radii. I've added property setters to allow overdensity radii to be
    #  set by external processes that might have measured a new value - for instance an iterative mass pipeline
    @property
    def r200(self) -> Quantity:
        """
        Getter for the radius at which the average density is 200 times the critical density.

        :return: The R200 in kpc.
        :rtype: Quantity
        """
        return self._r200

    @r200.setter
    def r200(self, new_value: Quantity):
        """
        The getter for the R200 property of the GalaxyCluster source class. This checks to make sure that the
        new value is an astropy Quantity, converts it to kpc, then updates all relevant attributes of this class.

        :param Quantity new_value:
        """
        # This checks that the input is a Quantity, then converts to kpc and to degrees
        new_value_kpc, new_value_deg = self._new_rad_checks(new_value)
        # For some reason these have to be set separately, stupid design on my part, but they are in different units
        #  so I guess I must have had some plan
        self._r200 = new_value_kpc
        self._radii['r200'] = new_value_deg

    @property
    def r500(self) -> Quantity:
        """
        Getter for the radius at which the average density is 500 times the critical density.

        :return: The R500 in kpc.
        :rtype: Quantity
        """
        return self._r500

    @r500.setter
    def r500(self, new_value: Quantity):
        """
        The getter for the R500 property of the GalaxyCluster source class. This checks to make sure that the
        new value is an astropy Quantity, converts it to kpc, then updates all relevant attributes of this class.

        :param Quantity new_value:
        """
        # This checks that the input is a Quantity, then converts to kpc and to degrees
        new_value_kpc, new_value_deg = self._new_rad_checks(new_value)
        # For some reason these have to be set separately, stupid design on my part, but they are in different units
        #  so I guess I must have had some plan
        self._r500 = new_value_kpc
        self._radii['r500'] = new_value_deg

    @property
    def r2500(self) -> Quantity:
        """
        Getter for the radius at which the average density is 2500 times the critical density.

        :return: The R2500 in kpc.
        :rtype: Quantity
        """
        return self._r2500

    @r2500.setter
    def r2500(self, new_value: Quantity):
        """
        The getter for the R2500 property of the GalaxyCluster source class. This checks to make sure that the
        new value is an astropy Quantity, converts it to kpc, then updates all relevant attributes of this class.

        :param Quantity new_value:
        """
        # This checks that the input is a Quantity, then converts to kpc and to degrees
        new_value_kpc, new_value_deg = self._new_rad_checks(new_value)
        # For some reason these have to be set separately, stupid design on my part, but they are in different units
        #  so I guess I must have had some plan
        self._r2500 = new_value_kpc
        self._radii['r2500'] = new_value_deg

    # Property getters for other observables I've allowed to be passed in.
    @property
    def weak_lensing_mass(self) -> Quantity:
        """
        Gets the weak lensing mass passed in at initialisation of the source.

        :return: Two quantities, the weak lensing mass, and the weak lensing mass error in Msun. If the
            values were not passed in at initialisation, the returned values will be None.
        :rtype: Quantity
        """
        if self._wl_mass is not None:
            wl_list = [self._wl_mass.value]
            wl_unit = self._wl_mass.unit
        else:
            wl_list = [np.NaN]
            wl_unit = ''

        if self._wl_mass_err is None:
            wl_list.append(np.NaN)
        elif isinstance(self._wl_mass_err, Quantity) and not self._wl_mass_err.isscalar:
            wl_list += list(self._wl_mass_err.value)
        elif isinstance(self._wl_mass_err, Quantity) and self._wl_mass_err.isscalar:
            wl_list.append(self._wl_mass_err.value)

        return Quantity(wl_list, wl_unit)

    @property
    def richness(self) -> Quantity:
        """
        Gets the richness passed in at initialisation of the source.

        :return: Two floats, the richness, and the richness error. If the values were not passed in at
            initialisation, the returned values will be None.
        :rtype: Quantity
        """
        if self._richness is not None:
            r_list = [self._richness]
        else:
            r_list = [np.NaN]

        if self._richness_err is None:
            r_list.append(np.NaN)
        elif isinstance(self._richness_err, (float, int)):
            r_list.append(self._richness_err)
        elif isinstance(self._richness_err, list):
            r_list += self._richness_err
        elif isinstance(self._richness_err, np.ndarray):
            r_list += list(self._richness_err)

        return Quantity(r_list)

    def get_results(self, outer_radius: Union[str, Quantity], model: str = 'constant*tbabs*apec',
                    inner_radius: Union[str, Quantity] = Quantity(0, 'arcsec'), par: str = None,
                    group_spec: bool = True, min_counts: int = 5, min_sn: float = None, over_sample: float = None):
        """
        Important method that will retrieve fit results from the source object. Either for a specific
        parameter of a given region-model combination, or for all of them. If a specific parameter is requested,
        all matching values from the fit will be returned in an N row, 3 column numpy array (column 0 is the value,
        column 1 is err-, and column 2 is err+). If no parameter is specified, the return will be a dictionary
        of such numpy arrays, with the keys corresponding to parameter names.

        This overrides the BaseSource method, but the only difference is that this has a default model, which
        is what single_temp_apec fits (constant*tbabs*apec).

        :param str/Quantity outer_radius: The name or value of the outer radius that was used for the generation of
            the spectra which were fitted to produce the desired result (for instance 'r200' would be acceptable
            for a GalaxyCluster, or Quantity(1000, 'kpc')). If 'region' is chosen (to use the regions in
            region files), then any inner radius will be ignored.
        :param str model: The name of the fitted model that you're requesting the results
            from (e.g. constant*tbabs*apec).
        :param str/Quantity inner_radius: The name or value of the inner radius that was used for the generation of
            the spectra which were fitted to produce the desired result (for instance 'r500' would be acceptable
            for a GalaxyCluster, or Quantity(300, 'kpc')). By default this is zero arcseconds, resulting in a
            circular spectrum.
        :param str par: The name of the parameter you want a result for.
        :param bool group_spec: Whether the spectra that were fitted for the desired result were grouped.
        :param float min_counts: The minimum counts per channel, if the spectra that were fitted for the
            desired result were grouped by minimum counts.
        :param float min_sn: The minimum signal to noise per channel, if the spectra that were fitted for the
            desired result were grouped by minimum signal to noise.
        :param float over_sample: The level of oversampling applied on the spectra that were fitted.
        :return: The requested result value, and uncertainties.
        """
        return super().get_results(outer_radius, model, inner_radius, par, group_spec, min_counts, min_sn, over_sample)

    def get_luminosities(self, outer_radius: Union[str, Quantity], model: str = 'constant*tbabs*apec',
                         inner_radius: Union[str, Quantity] = Quantity(0, 'arcsec'), lo_en: Quantity = None,
                         hi_en: Quantity = None, group_spec: bool = True, min_counts: int = 5, min_sn: float = None,
                         over_sample: float = None):
        """
        Get method for luminosities calculated from model fits to spectra associated with this source.
        Either for given energy limits (that must have been specified when the fit was first performed), or
        for all luminosities associated with that model. Luminosities are returned as a 3 column numpy array;
        the 0th column is the value, the 1st column is the err-, and the 2nd is err+.

        This overrides the BaseSource method, but the only difference is that this has a default model, which
        is what single_temp_apec fits (constant*tbabs*apec).

        :param str/Quantity outer_radius: The name or value of the outer radius that was used for the generation of
            the spectra which were fitted to produce the desired result (for instance 'r200' would be acceptable
            for a GalaxyCluster, or Quantity(1000, 'kpc')). If 'region' is chosen (to use the regions in
            region files), then any inner radius will be ignored.
        :param str model: The name of the fitted model that you're requesting the luminosities
            from (e.g. constant*tbabs*apec).
        :param str/Quantity inner_radius: The name or value of the inner radius that was used for the generation of
            the spectra which were fitted to produce the desired result (for instance 'r500' would be acceptable
            for a GalaxyCluster, or Quantity(300, 'kpc')). By default this is zero arcseconds, resulting in a
            circular spectrum.
        :param Quantity lo_en: The lower energy limit for the desired luminosity measurement.
        :param Quantity hi_en: The upper energy limit for the desired luminosity measurement.
        :param bool group_spec: Whether the spectra that were fitted for the desired result were grouped.
        :param float min_counts: The minimum counts per channel, if the spectra that were fitted for the
            desired result were grouped by minimum counts.
        :param float min_sn: The minimum signal to noise per channel, if the spectra that were fitted for the
            desired result were grouped by minimum signal to noise.
        :param float over_sample: The level of oversampling applied on the spectra that were fitted.
        :return: The requested luminosity value, and uncertainties.
        """
        return super().get_luminosities(outer_radius, model, inner_radius, lo_en, hi_en, group_spec, min_counts,
                                        min_sn, over_sample)

    # This does duplicate some of the functionality of get_results, but in a more specific way. I think its
    #  justified considering how often the cluster temperature is used in X-ray cluster studies.
    def get_temperature(self, outer_radius: Union[str, Quantity], model: str = 'constant*tbabs*apec',
                        inner_radius: Union[str, Quantity] = Quantity(0, 'arcsec'), group_spec: bool = True,
                        min_counts: int = 5, min_sn: float = None, over_sample: float = None):
        """
        Convenience method that calls get_results to retrieve temperature measurements. All matching values
        from the fit will be returned in an N row, 3 column numpy array (column 0 is the value,
        column 1 is err-, and column 2 is err+).

        :param str/Quantity outer_radius: The name or value of the outer radius that was used for the generation of
            the spectra which were fitted to produce the desired result (for instance 'r200' would be acceptable
            for a GalaxyCluster, or Quantity(1000, 'kpc')). If 'region' is chosen (to use the regions in
            region files), then any inner radius will be ignored.
        :param str model: The name of the fitted model that you're requesting the results
            from (e.g. constant*tbabs*apec).
        :param str/Quantity inner_radius: The name or value of the inner radius that was used for the generation of
            the spectra which were fitted to produce the desired result (for instance 'r500' would be acceptable
            for a GalaxyCluster, or Quantity(300, 'kpc')). By default this is zero arcseconds, resulting in a
            circular spectrum.
        :param bool group_spec: Whether the spectra that were fitted for the desired result were grouped.
        :param float min_counts: The minimum counts per channel, if the spectra that were fitted for the
            desired result were grouped by minimum counts.
        :param float min_sn: The minimum signal to noise per channel, if the spectra that were fitted for the
            desired result were grouped by minimum signal to noise.
        :param float over_sample: The level of oversampling applied on the spectra that were fitted.
        :return: The temperature value, and uncertainties.
        """
        res = self.get_results(outer_radius, model, inner_radius, "kT", group_spec, min_counts, min_sn, over_sample)

        return Quantity(res, 'keV')

    def _get_spec_based_profiles(self, search_key: str, radii: Quantity = None, group_spec: bool = True,
                                 min_counts: int = 5, min_sn: float = None, over_sample: float = None,
                                 set_id: int = None) -> Union[BaseProfile1D, List[BaseProfile1D]]:
        """
        The generic get method for profiles which have been based on spectra, the only thing that tends to change
        about how we search for them is the specific search key. Largely copied from get_annular_spectra.

        :param str search_key: The profile search key, e.g. combined_1d_proj_temperature_profile.
        :param Quantity radii: The annulus boundary radii that were used to generate the annular spectra set
            from which the projected temperature profile was measured.
        :param bool group_spec: Was the spectrum set used to generate the profile grouped
        :param float min_counts: If the spectrum set used to generate the profile was grouped on minimum
            counts, what was the minimum number of counts?
        :param float min_sn: If the spectrum set used to generate the profile was grouped on minimum signal to
            noise, what was the minimum signal to noise.
        :param float over_sample: If the spectrum set used to generate the profile was over sampled, what was
            the level of over sampling used?
        :param int set_id: The unique identifier of the annular spectrum set used to generate the profile.
            Passing a value for this parameter will override any other information that you have given this method.
        :return: An XGA profile object if there is an exact match, a list of such objects if there are multiple matches.
        :rtype: Union[BaseProfile1D, List[BaseProfile1D]]
        """
        if group_spec and min_counts is not None:
            extra_name = "_mincnt{}".format(min_counts)
        elif group_spec and min_sn is not None:
            extra_name = "_minsn{}".format(min_sn)
        else:
            extra_name = ''

        # And if it was oversampled during generation then we need to include that as well
        if over_sample is not None:
            extra_name += "_ovsamp{ov}".format(ov=over_sample)

        # Combines the annular radii into a string, and makes sure the radii are in degrees, as radii are in
        #  degrees in the storage key
        if radii is not None:
            # We're dealing with the best case here, the user has passed radii, so we can generate an exact
            #  storage key and look for a single match
            ann_rad_str = "_".join(self.convert_radius(radii, 'deg').value.astype(str))
            spec_storage_name = "ra{ra}_dec{dec}_ar{ar}_grp{gr}"
            spec_storage_name = spec_storage_name.format(ra=self.default_coord[0].value,
                                                         dec=self.default_coord[1].value,
                                                         ar=ann_rad_str, gr=group_spec)
            spec_storage_name += extra_name
        else:
            # This is a worse case, we don't have radii, so we split the known parts of the key into a list
            #  and we'll look for partial matches
            pos_str = "ra{ra}_dec{dec}".format(ra=self.default_coord[0].value, dec=self.default_coord[1].value)
            grp_str = "grp{gr}".format(gr=group_spec) + extra_name
            spec_storage_name = [pos_str, grp_str]

        # If the user hasn't passed a set ID AND the user has passed radii then we'll go looking with out
        #  properly constructed storage key
        if set_id is None and radii is not None:
            matched_prods = self.get_products(search_key, extra_key=spec_storage_name)
        # But if the user hasn't passed an ID AND the radii are None then we look for partial matches
        elif set_id is None and radii is None:
            matched_prods = [p for p in self.get_products(search_key)
                             if spec_storage_name[0] in p.storage_key and spec_storage_name[1] in p.storage_key]
        # However if they have passed a setID then this over-rides everything else
        else:
            matched_prods = [p for p in self.get_products(search_key) if p.set_ident == set_id]

        return matched_prods

    def get_proj_temp_profiles(self, radii: Quantity = None, group_spec: bool = True, min_counts: int = 5,
                               min_sn: float = None, over_sample: float = None, set_id: int = None) \
            -> Union[ProjectedGasTemperature1D, List[ProjectedGasTemperature1D]]:
        """
        A get method for projected temperature profiles generated by XGA's XSPEC interface. This works identically
        to the get_annular_spectra method, because projected temperature profiles are generated from annular spectra,
        and as such can be described by the same parameters.

        :param Quantity radii: The annulus boundary radii that were used to generate the annular spectra set
            from which the projected temperature profile was measured.
        :param bool group_spec: Was the spectrum set used to generate the profile grouped
        :param float min_counts: If the spectrum set used to generate the profile was grouped on minimum
            counts, what was the minimum number of counts?
        :param float min_sn: If the spectrum set used to generate the profile was grouped on minimum signal to
            noise, what was the minimum signal to noise.
        :param float over_sample: If the spectrum set used to generate the profile was over sampled, what was
            the level of over sampling used?
        :param int set_id: The unique identifier of the annular spectrum set used to generate the profile.
            Passing a value for this parameter will override any other information that you have given this method.
        :return: An XGA ProjectedGasTemperature1D object if there is an exact match, a list of such objects
            if there are multiple matches.
        :rtype: Union[ProjectedGasTemperature1D, List[ProjectedGasTemperature1D]]
        """
        matched_prods = self._get_spec_based_profiles("combined_1d_proj_temperature_profile", radii, group_spec,
                                                      min_counts, min_sn, over_sample, set_id)

        if len(matched_prods) == 1:
            matched_prods = matched_prods[0]
        elif len(matched_prods) == 0:
            raise NoProductAvailableError("No matching 1D projected temperature profiles can be found.")

        return matched_prods

    def get_3d_temp_profiles(self, radii: Quantity = None, group_spec: bool = True, min_counts: int = 5,
                             min_sn: float = None, over_sample: float = None, set_id: int = None) \
            -> Union[GasTemperature3D, List[GasTemperature3D]]:
        """
        A get method for 3D temperature profiles generated by XGA's de-projection routines.

        :param Quantity radii: The annulus boundary radii that were used to generate the annular spectra set
            from which the projected temperature profile was measured.
        :param bool group_spec: Was the spectrum set used to generate the profile grouped
        :param float min_counts: If the spectrum set used to generate the profile was grouped on minimum
            counts, what was the minimum number of counts?
        :param float min_sn: If the spectrum set used to generate the profile was grouped on minimum signal to
            noise, what was the minimum signal to noise.
        :param float over_sample: If the spectrum set used to generate the profile was over sampled, what was
            the level of over sampling used?
        :param int set_id: The unique identifier of the annular spectrum set used to generate the profile.
            Passing a value for this parameter will override any other information that you have given this method.
        :return: An XGA ProjectedGasTemperature1D object if there is an exact match, a list of such objects
            if there are multiple matches.
        :rtype: Union[ProjectedGasTemperature1D, List[ProjectedGasTemperature1D]]
        """
        matched_prods = self._get_spec_based_profiles("combined_gas_temperature_profile", radii, group_spec,
                                                      min_counts, min_sn, over_sample, set_id)

        if len(matched_prods) == 1:
            matched_prods = matched_prods[0]
        elif len(matched_prods) == 0:
            raise NoProductAvailableError("No matching 3D temperature profiles can be found.")

        return matched_prods

    def get_apec_norm_profiles(self, radii: Quantity = None, group_spec: bool = True, min_counts: int = 5,
                               min_sn: float = None, over_sample: float = None, set_id: int = None) \
            -> Union[APECNormalisation1D, List[APECNormalisation1D]]:
        """
        A get method for APEC normalisation profiles generated by XGA's XSPEC interface.

        :param Quantity radii: The annulus boundary radii that were used to generate the annular spectra set
            from which the normalisation profile was measured.
        :param bool group_spec: Was the spectrum set used to generate the profile grouped
        :param float min_counts: If the spectrum set used to generate the profile was grouped on minimum
            counts, what was the minimum number of counts?
        :param float min_sn: If the spectrum set used to generate the profile was grouped on minimum signal to
            noise, what was the minimum signal to noise.
        :param float over_sample: If the spectrum set used to generate the profile was over sampled, what was
            the level of over sampling used?
        :param int set_id: The unique identifier of the annular spectrum set used to generate the profile.
            Passing a value for this parameter will override any other information that you have given this method.
        :return: An XGA APECNormalisation1D object if there is an exact match, a list of such objects
            if there are multiple matches.
        :rtype: Union[ProjectedGasTemperature1D, List[ProjectedGasTemperature1D]]
        """
        matched_prods = self._get_spec_based_profiles("combined_1d_apec_norm_profile", radii, group_spec,
                                                      min_counts, min_sn, over_sample, set_id)

        if len(matched_prods) == 1:
            matched_prods = matched_prods[0]
        elif len(matched_prods) == 0:
            raise NoProductAvailableError("No matching APEC normalisation profiles can be found.")

        return matched_prods

    def get_density_profiles(self, outer_rad: Union[Quantity, str] = None, method: str = None, obs_id: str = None,
                             inst: str = None, central_coord: Quantity = None, radii: Quantity = None,
                             pix_step: int = 1, min_snr: Union[float, int] = 0.0, psf_corr: bool = True,
                             psf_model: str = "ELLBETA", psf_bins: int = 4, psf_algo: str = "rl", psf_iter: int = 15,
                             group_spec: bool = True, min_counts: int = 5, min_sn: float = None,
                             over_sample: float = None, set_id: int = None) -> Union[GasDensity3D, List[GasDensity3D]]:
        """
        This is a get method for density profiles generated by XGA, both using surface brightness profiles and spectra.
        Having to account for two different methods is why this get method has so many arguments that can be passed. If
        multiple matches for the passed variables are found, then a list of density profiles will be returned,
        otherwise only a single profile will be returned.

        :param Quantity/str outer_rad: The outer radius of the density profile, either as a name ('r500' for instance)
            or an astropy Quantity.
        :param str method: The method used to generate the density profile. For a profile created by fitting a model
            to a surface brightness profile this should be the name of the model, for a profile from annular spectra
            this should be 'spec', and for a profile generated directly from the data of a surface brightness profile
            this should be 'onion'.
        :param str obs_id: The ObsID used to generate the profile in question, default is None (which will search for
            profiles generated from combined data).
        :param str inst: The instrument used to generate the profile in question, default is None (which will
            search for profiles generated from combined data).
        :param Quantity central_coord: The central coordinate of the density profile. Default is None, which means
            we shall use the default coordinate of this source.
        :param Quantity radii: If known, the radii that were used to measure the density profile.
        :param int pix_step: The width of each annulus in pixels used to generate the profile, for profiles based on
            surface brightness.
        :param float min_snr: The minimum signal to noise imposed upon the profile, for profiles based on
            surface brightness.
        :param bool psf_corr: Is the brightness profile corrected for PSF effects, for profiles based on
            surface brightness.
        :param str psf_model: If PSF corrected, the PSF model used, for profiles based on surface brightness.
        :param int psf_bins: If PSF corrected, the number of bins per side, for profiles based on surface brightness.
        :param str psf_algo: If PSF corrected, the algorithm used, for profiles based on surface brightness.
        :param int psf_iter: If PSF corrected, the number of algorithm iterations, for profiles based on
            surface brightness.
        :param bool group_spec: Was the spectrum set used to generate the profile grouped.
        :param float min_counts: If the spectrum set used to generate the profile was grouped on minimum
            counts, what was the minimum number of counts?
        :param float min_sn: If the spectrum set used to generate the profile was grouped on minimum signal to
            noise, what was the minimum signal to noise.
        :param float over_sample: If the spectrum set used to generate the profile was over sampled, what was
            the level of over sampling used?
        :param int set_id: The unique identifier of the annular spectrum set used to generate the profile.
            Passing a value for this parameter will override any other information that you have given this method.
        :return:
        :rtype: Union[GasDensity3D, List[GasDensity3D]]
        """
        if outer_rad is not None and isinstance(outer_rad, str):
            outer_rad = self.get_radius(outer_rad, 'deg')
        elif outer_rad is not None and isinstance(outer_rad, Quantity):
            outer_rad = self.convert_radius(outer_rad, 'deg')
        elif outer_rad is not None:
            raise ValueError("Outer radius may only be a string or an astropy quantity")

        if (obs_id == "combined" or inst == "combined" or obs_id is None or inst is None) and method != 'spec':
            interim_prods = self.get_combined_profiles("gas_density", central_coord, radii)
        elif method != 'spec':
            interim_prods = self.get_profiles("gas_density", obs_id, inst, central_coord, radii)
        elif (obs_id == "combined" or inst == "combined" or obs_id is None or inst is None) and method == 'spec':
            interim_prods = self._get_spec_based_profiles('combined_gas_density_profile', radii, group_spec, min_counts,
                                                          min_sn, over_sample, set_id)
        else:
            interim_prods = self._get_spec_based_profiles('gas_density_profile', radii, group_spec, min_counts, min_sn,
                                                          over_sample, set_id)

        # The methods I used to get this far will already have gotten upset if there are no matches, so I don't need
        #  to check they exist, but I do need to check if I have a list or a single object
        if not isinstance(interim_prods, list):
            interim_prods = [interim_prods]

        matched_prods = []
        if method != 'spec' and any([p.density_method == method for p in interim_prods]):
            interim_prods = [p for p in interim_prods if p.density_method == method]
            for dens_prof in interim_prods:
                p = dens_prof.generation_profile
                if not psf_corr and p.pix_step == pix_step and p.min_snr == min_snr and p.psf_corrected == psf_corr:
                    matched_prods.append(dens_prof)
                elif psf_corr and p.pix_step == pix_step and p.min_snr == min_snr and p.psf_corrected == psf_corr and \
                        p.psf_model == psf_model and p.psf_bins == psf_bins and p.psf_algorithm == psf_algo \
                        and p.psf_iterations == psf_iter:
                    matched_prods.append(dens_prof)

        elif method == 'spec' and any([p.density_method == 'spec' for p in interim_prods]):
            matched_prods = interim_prods
        else:
            matched_prods = interim_prods

        if outer_rad is not None:
            matched_prods = [im for im in matched_prods if self.convert_radius(im.outer_radius, 'deg') == outer_rad]

        if len(matched_prods) == 1:
            matched_prods = matched_prods[0]
        elif len(matched_prods) == 0:
            raise NoProductAvailableError("Cannot find any density profiles matching your input.")

        return matched_prods

    def get_hydrostatic_mass_profiles(self, temp_prof: GasTemperature3D = None, temp_model_name: str = None,
                                      dens_prof: GasDensity3D = None, dens_model_name: str = None,
                                      radii: Quantity = None) -> Union[HydrostaticMass, List[HydrostaticMass]]:
        """
        A get method for hydrostatic mass profiles associated with this galaxy cluster. This works in a slightly
        different way to the temperature and density profile get methods, as you can pass the gas temperature and
        density profiles used to generate a hydrostatic mass profile to find it. If none of the optional
        arguments are passed then all hydrostatic mass profiles associated with this source will be returned, if
        only some are passed then mass profiles which match the limited information will be found.

        :param GasTemperature3D temp_prof: The temperature profile used to generate the required hydrostatic mass
            profile, default is None.
        :param str temp_model_name: The name of the model used to fit the temperature profile used to generate the
            required hydrostatic mass profile, default is None.
        :param GasDensity3D dens_prof: The density profile used to generate the required hydrostatic mass
            profile, default is None.
        :param str dens_model_name: The name of the model used to fit the density profile used to generate the
            required hydrostatic mass profile, default is None.
        :param Quantity radii: The radii at which the hydrostatic mass profile was measured, default is None.
        :return: Either a single hydrostatic mass profile, when there is a unique match, or a list of hydrostatic
            mass profiles if there is not.
        :rtype: Union[HydrostaticMass, List[HydrostaticMass]]
        """
        # Get all the hydrostatic mass profiles associated with this source
        matched_prods = self.get_profiles('combined_hydrostatic_mass')

        # Convert the radii to degrees for comparison with deg radii later
        if radii is not None:
            radii = self.convert_radius(radii, 'deg')

        # Checking steps, looking for matches with the information passed by the user.
        if temp_prof is not None:
            matched_prods = [p for p in matched_prods if p.temperature_profile != temp_prof]

        if dens_prof is not None:
            matched_prods = [p for p in matched_prods if p.density_profile != dens_prof]

        if temp_model_name is not None:
            matched_prods = [p for p in matched_prods if p.temperature_model.name != temp_model_name]

        if dens_model_name is not None:
            matched_prods = [p for p in matched_prods if p.density_model.name != dens_model_name]

        if radii is not None:
            matched_prods = [p for p in matched_prods if p.deg_radii != radii]

        if isinstance(matched_prods, list) and len(matched_prods) == 0:
            raise NoProductAvailableError("No matching hydrostatic mass profiles can be found.")
        return matched_prods

    def view_brightness_profile(self, reg_type: str, central_coord: Quantity = None, pix_step: int = 1,
                                min_snr: Union[float, int] = 0.0, figsize: tuple = (10, 7), xscale: str = 'log',
                                yscale: str = 'log', back_sub: bool = True, lo_en: Quantity = Quantity(0.5, 'keV'),
                                hi_en: Quantity = Quantity(2.0, 'keV')):
        """
        A method that generates and displays brightness profile objects for this galaxy cluster. Interloper
        sources are excluded, and any fits performed to pre-existing brightness profiles which are being
        viewed will also be displayed. The profile will be generated using a RateMap between the energy bounds
        specified by lo_en and hi_en.

        :param str reg_type: The region in which to view the radial brightness profile.
        :param Quantity central_coord: The central coordinate of the brightness profile.
        :param int pix_step: The width (in pixels) of each annular bin, default is 1.
        :param float/int min_snr: The minimum signal to noise allowed for each radial bin. This is 0 by
            default, which disables any automatic re-binning.
        :param tuple figsize: The desired size of the figure, the default is (10, 7)
        :param str xscale: The scaling to be applied to the x axis, default is log.
        :param str yscale: The scaling to be applied to the y axis, default is log.
        :param bool back_sub: Should the plotted data be background subtracted, default is True.
        :param Quantity lo_en: The lower energy bound of the RateMap to generate the profile from.
        :param Quantity hi_en: The upper energy bound of the RateMap to generate the profile from.
        """
        allowed_rtype = ["custom", "r500", "r200", "r2500"]
        if reg_type not in allowed_rtype:
            raise ValueError("The only allowed region types are {}".format(", ".join(allowed_rtype)))

        # Check that the valid region choice actually has an entry that is not None
        if reg_type == "custom" and self._custom_region_radius is None:
            raise NoRegionsError("No custom region has been setup for this cluster")
        elif reg_type == "r200" and self._r200 is None:
            raise NoRegionsError("No R200 region has been setup for this cluster")
        elif reg_type == "r500" and self._r500 is None:
            raise NoRegionsError("No R500 region has been setup for this cluster")
        elif reg_type == "r2500" and self._r2500 is None:
            raise NoRegionsError("No R2500 region has been setup for this cluster")

        comb_rt = self.get_combined_ratemaps(lo_en, hi_en)
        # If there have been PSF deconvolutions of the above data, then we can grab them too
        # I still do it this way rather than with get_combined_ratemaps because I want ALL PSF corrected ratemaps
        en_key = "bound_{l}-{u}".format(l=lo_en.value, u=hi_en.value)
        psf_comb_rts = [rt for rt in self.get_products("combined_ratemap", just_obj=False)
                        if en_key + "_" in rt[-2]]

        # Fetch the mask that will remove all interloper sources from the combined ratemap
        int_mask = self.get_interloper_mask()

        if central_coord is None:
            central_coord = self.default_coord

        # Read out the radii
        rad = self.get_radius(reg_type)

        # This fetches any profiles that might have already been generated to our required specifications
        try:
            sb_profile = self.get_1d_brightness_profile(rad, pix_step=pix_step, min_snr=min_snr, lo_en=lo_en,
                                                        hi_en=hi_en)
            if isinstance(sb_profile, list):
                raise ValueError("There are multiple matches for this brightness profile, and its the developers "
                                 "fault not yours.")
        except NoProductAvailableError:
            sb_profile, success = radial_brightness(comb_rt, central_coord, rad, self._back_inn_factor,
                                                    self._back_out_factor, int_mask, self.redshift, pix_step, kpc,
                                                    self.cosmo, min_snr)
            self.update_products(sb_profile)

        for psf_comb_rt in psf_comb_rts:
            p_rt = psf_comb_rt[-1]

            try:
                psf_sb_profile = self.get_1d_brightness_profile(rad, pix_step=pix_step, min_snr=min_snr,
                                                                psf_corr=True, psf_model=p_rt.psf_model,
                                                                psf_bins=p_rt.psf_bins, psf_algo=p_rt.psf_algorithm,
                                                                psf_iter=p_rt.psf_iterations, lo_en=lo_en, hi_en=hi_en)
                if isinstance(psf_sb_profile, list):
                    raise ValueError("There are multiple matches for this brightness profile, and its the developers "
                                     "fault not yours.")
            except NoProductAvailableError:
                psf_sb_profile, success = radial_brightness(psf_comb_rt[-1], central_coord, rad,
                                                            self._back_inn_factor, self._back_out_factor, int_mask,
                                                            self.redshift, pix_step, kpc, self.cosmo, min_snr)
                self.update_products(psf_sb_profile)

            sb_profile += psf_sb_profile

        draw_rads = {}
        for r_name in self._radii:
            if r_name not in ['search', 'custom']:
                new_key = "R$_{" + r_name[1:] + "}$"
                draw_rads[new_key] = self.get_radius(r_name, sb_profile.radii_unit)
            elif r_name == "custom":
                draw_rads["Custom"] = self.get_radius(r_name, sb_profile.radii_unit)

        sb_profile.view(xscale=xscale, yscale=yscale, figsize=figsize, draw_rads=draw_rads, back_sub=back_sub)

    def combined_lum_conv_factor(self, outer_radius: Union[str, Quantity], lo_en: Quantity, hi_en: Quantity,
                                 inner_radius: Union[str, Quantity] = Quantity(0, 'arcsec'), group_spec: bool = True,
                                 min_counts: int = 5, min_sn: float = None, over_sample: float = None) -> Quantity:
        """
        Combines conversion factors calculated for this source with individual instrument-observation
        spectra, into one overall conversion factor.

        :param str/Quantity outer_radius: The name or value of the outer radius of the spectra that should be used
            to calculate conversion factors (for instance 'r200' would be acceptable for a GalaxyCluster, or
            Quantity(1000, 'kpc')). If 'region' is chosen (to use the regions in region files), then any
            inner radius will be ignored.
        :param str/Quantity inner_radius: The name or value of the inner radius of the spectra that should be used
            to calculate conversion factors (for instance 'r500' would be acceptable for a GalaxyCluster, or
            Quantity(300, 'kpc')). By default this is zero arcseconds, resulting in a circular spectrum.
        :param Quantity lo_en: The lower energy limit of the conversion factors.
        :param Quantity hi_en: The upper energy limit of the conversion factors.
        :param bool group_spec: Whether the spectra that were used for fakeit were grouped.
        :param float min_counts: The minimum counts per channel, if the spectra that were used for fakeit
            were grouped by minimum counts.
        :param float min_sn: The minimum signal to noise per channel, if the spectra that were used for fakeit
            were grouped by minimum signal to noise.
        :param float over_sample: The level of oversampling applied on the spectra that were used for fakeit.
        :return: A combined conversion factor that can be applied to a combined ratemap to calculate luminosity.
        :rtype: Quantity
        """
        # Grabbing the relevant spectra
        spec = self.get_spectra(outer_radius, inner_radius=inner_radius, group_spec=group_spec, min_counts=min_counts,
                                min_sn=min_sn, over_sample=over_sample)
        # Setting up variables to be added into
        av_lum = Quantity(0, "erg/s")
        total_phot = 0

        if isinstance(spec, Spectrum):
            spec = [spec]

        # Cycling through the relevant spectra
        for s in spec:
            # The luminosity is added to the average luminosity variable, will be divided by N
            #  spectra at the end.
            av_lum += s.get_conv_factor(lo_en, hi_en, "tbabs*apec")[1]
            # Multiplying by 1e+4 because that is the length of the simulated exposure in seconds
            total_phot += s.get_conv_factor(lo_en, hi_en, "tbabs*apec")[2] * 1e+4

        # Then the total combined rate is the total number of photons / the total summed exposure (which
        #  is just 10000 seconds multiplied by the number of spectra).
        total_rate = Quantity(total_phot / (1e+4 * len(spec)), 'ct/s')

        # Making av_lum actually an average
        av_lum /= len(spec)

        # Calculating and returning the combined factor.
        return av_lum / total_rate

    def norm_conv_factor(self, outer_radius: Union[str, Quantity], lo_en: Quantity, hi_en: Quantity,
                         inner_radius: Union[str, Quantity] = Quantity(0, 'arcsec'), group_spec: bool = True,
                         min_counts: int = 5, min_sn: float = None, over_sample: float = None, obs_id: str = None,
                         inst: str = None) -> Quantity:
        """
        Combines count-rate to normalisation conversion factors associated with this source.

        :param str/Quantity outer_radius: The name or value of the outer radius of the spectra that should be used
            to calculate conversion factors (for instance 'r200' would be acceptable for a GalaxyCluster, or
            Quantity(1000, 'kpc')). If 'region' is chosen (to use the regions in region files), then any
            inner radius will be ignored.
        :param str/Quantity inner_radius: The name or value of the inner radius of the spectra that should be used
            to calculate conversion factors (for instance 'r500' would be acceptable for a GalaxyCluster, or
            Quantity(300, 'kpc')). By default this is zero arcseconds, resulting in a circular spectrum.
        :param Quantity lo_en: The lower energy limit of the conversion factors.
        :param Quantity hi_en: The upper energy limit of the conversion factors.
        :param bool group_spec: Whether the spectra that were used for fakeit were grouped.
        :param float min_counts: The minimum counts per channel, if the spectra that were used for fakeit
            were grouped by minimum counts.
        :param float min_sn: The minimum signal to noise per channel, if the spectra that were used for fakeit
            were grouped by minimum signal to noise.
        :param float over_sample: The level of oversampling applied on the spectra that were used for fakeit.
        :param str obs_id: The ObsID to fetch a conversion factor for, default is None which means the combined
            conversion factor will be returned.
        :param str inst: The instrument to fetch a conversion factor for, default is None which means the combined
            conversion factor will be returned.
        :return: A combined conversion factor that can be applied to a combined ratemap to calculate luminosity.
        :rtype: Quantity
        """
        # Check the ObsID and instrument inputs
        if all([obs_id is None, inst is None]):
            pass
        elif all([obs_id is not None, inst is not None]):
            pass
        else:
            raise ValueError("If a value is supplied for obs_id, then a value must be supplied for inst as well, and "
                             "vice versa.")

        # Grabbing the relevant spectra
        spec = self.get_spectra(outer_radius, inner_radius=inner_radius, group_spec=group_spec, min_counts=min_counts,
                                min_sn=min_sn, over_sample=over_sample, obs_id=obs_id, inst=inst)

        # Its just easier if we know that the spectra are in a list
        if isinstance(spec, Spectrum):
            spec = [spec]

        mean_areas = []
        rates = []
        for s in spec:
            s: Spectrum

            # For the current spectrum we retrieve the ARF information so that we can use it to weight things with
            #  later
            ens = (s.eff_area_hi_en + s.eff_area_lo_en) / 2
            ars = s.eff_area

            # This finds only the areas in the current energy range we're considering
            rel_ars = ars[np.argwhere((ens <= hi_en) & (ens >= lo_en)).T[0]]
            # And finds the mean effective area in that range
            mean_areas.append(rel_ars.mean().value)

            # Then we fetch the count rate for the fakeit run of the current spectrum
            rates.append(s.get_conv_factor(lo_en, hi_en, "tbabs*apec")[2].value)

        # Just putting rates as an array for convenience
        rates = np.array(rates)
        # These normalisation factors put all the conversion factors on the same footing, by weighting by
        #  effective area of the spectra
        norm_factors = np.array(mean_areas)/mean_areas[0]
        # The total photons is the sum of the rates multiplied by the simulation exposure (always 1e+4 seconds)
        #  weighted by the effective area.
        total_phot = (rates * (1e+4/norm_factors)).sum()

        # Then the total combined rate is the total number of photons / the total summed exposure, which again is a
        #  sum of the simulation exposure weighted by the effective areas of all the spectra.
        total_rate = Quantity(total_phot / (1e+4/norm_factors).sum(), 'ct/s')

        # Then we return 1/the rate because this method calculates the conversion factor from count rate to
        #  normalisation (which is always 1 for these simulated spectra). The normalisation has units of cmm^-5,
        #  as is easily derived from the normalisation expression here
        #  (https://heasarc.gsfc.nasa.gov/xanadu/xspec/manual/XSmodelApec.html)
        return Quantity(1, 'cm^-5') / total_rate






#  This code is a part of X-ray: Generate and Analyse (XGA), a module designed for the XMM Cluster Survey (XCS).
#  Last modified by David J Turner (turne540@msu.edu) 25/11/2024, 09:53. Copyright (c) The Contributors

from typing import Union, List

from astropy.units import Quantity, Unit, UnitConversionError, kpc

from .base import BaseModel1D
from ..utils import r200, r500, r2500


class SimpleGNFWThermalPressure(BaseModel1D):
    """
    A model to fit galaxy cluster radial thermal pressure profiles, based on the generalized NFW profile equation
    proposed by https://ui.adsabs.harvard.edu/abs/2007ApJ...668....1N/abstract and used
    in https://ui.adsabs.harvard.edu/abs/2010A%26A...517A..92A/abstract - this one has had several parameters removed
    or frozen.

    :param Unit/str x_unit: The unit of the x-axis of this model, kpc for instance. May be passed as a string
        representation or an astropy unit object.
    :param Unit/str y_unit: The unit of the output of this model, keV/cm^3 for instance. May be passed as a string
        representation or an astropy unit object.
    :param List[Quantity] cust_start_pars: The start values of the model parameters for any fitting function that
        used start values. The units are checked against default start values.
    """
    def __init__(self, x_unit: Union[str, Unit] = 'kpc', y_unit: Union[str, Unit] = Unit('keV cm^-3'),
                 cust_start_pars: List[Quantity] = None):
        """
        The init of a subclass of the XGA BaseModel1D class, describing a model of how thermal pressure changes
        with radius.
        """
        # If a string representation of a unit was passed then we make it an astropy unit
        if isinstance(x_unit, str):
            x_unit = Unit(x_unit)
        if isinstance(y_unit, str):
            y_unit = Unit(y_unit)

        poss_y_units = [Unit('keV cm^-3')]
        y_convertible = [u.is_equivalent(y_unit) for u in poss_y_units]
        if not any(y_convertible):
            allowed = ", ".join([u.to_string() for u in poss_y_units])
            raise UnitConversionError("{p} is not convertible to any of the allowed units; "
                                      "{a}".format(p=y_unit.to_string(), a=allowed))
        else:
            yu_ind = y_convertible.index(True)

        poss_x_units = [kpc]
        x_convertible = [u.is_equivalent(x_unit) for u in poss_x_units]
        if not any(x_convertible):
            allowed = ", ".join([u.to_string() for u in poss_x_units])
            raise UnitConversionError("{p} is not convertible to any of the allowed units; "
                                      "{a}".format(p=x_unit.to_string(), a=allowed))
        else:
            xu_ind = x_convertible.index(True)

        p_zero_starts = [Quantity(0.1, 'keV cm^-3')]
        r_scale_starts = [Quantity(100, 'kpc'), Quantity(0.2, 'deg'), Quantity(0.05, r200), Quantity(0.1, r500),
                          Quantity(0.5, r2500)]
        alpha = Quantity(1, '')
        beta = Quantity(1, '')

        start_pars = [p_zero_starts[yu_ind], r_scale_starts[xu_ind], alpha, beta]

        if cust_start_pars is not None:
            # If the custom start parameters can run this gauntlet without tripping an error then we're all good
            # This method also returns the custom start pars converted to exactly the same units as the default
            start_pars = self.compare_units(cust_start_pars, start_pars)

        pz_priors = [{'prior': Quantity([0.0001, 10], 'keV cm^-3'), 'type': 'uniform'}]
        r_scale_priors = [{'prior': Quantity([0, 2000], 'kpc'), 'type': 'uniform'},
                         {'prior': Quantity([0, 1], 'deg'), 'type': 'uniform'},
                         {'prior': Quantity([0, 1], r200), 'type': 'uniform'},
                         {'prior': Quantity([0, 1], r500), 'type': 'uniform'},
                         {'prior': Quantity([0, 1], r2500), 'type': 'uniform'}]
        alpha_priors = {'prior': Quantity([0, 10], ''), 'type': 'uniform'}
        beta_priors = {'prior': Quantity([0, 10], ''), 'type': 'uniform'}

        priors = [pz_priors[yu_ind], r_scale_priors[xu_ind], alpha_priors, beta_priors]

        nice_pars = [r"P$_{0}$", r"$R_{s}$", r"$\alpha$", r"$\beta$"]
        info_dict = {'author': 'Arnaud et al.', 'year': 2010,
                     'reference': 'https://ui.adsabs.harvard.edu/abs/2007ApJ...668....1N/abstract',
                     'general': "A model to fit galaxy cluster radial thermal pressure profiles, \n"
                                "based on the generalized NFW profile equation, but with gamma fixed."}

        super().__init__(x_unit, y_unit, start_pars, priors, 'simple_gnfw_pressure',
                         'Simplified gNFW Pressure Profile', nice_pars, 'Thermal Pressure', info_dict)

    @staticmethod
    def model(x: Quantity, p_zero: Quantity, r_scale: Quantity, alpha: Quantity, beta: Quantity) -> Quantity:
        """
        The model function for the simplified generalized NFW pressure profile.

        :param Quantity x: The radii to calculate y values for.
        :param Quantity p_zero: The pressure normalization for the model.
        :param Quantity r_scale: The scale radius for the radial pressure profile.
        :param Quantity alpha: The alpha slope parameter, for the intermediate slope.
        :param Quantity beta: The beta slope parameter, for the outer slope.
        :return: The y values corresponding to the input x values.
        :rtype: Quantity
        """
        gamma = 0.31
        rad_rat = x/r_scale

        result = p_zero * (1/(rad_rat**gamma * (1 + rad_rat**alpha)**((beta - gamma)/alpha)))

        return result


class GNFWThermalPressure(BaseModel1D):
    """
    The full generalized NFW profile equation used to fit galaxy cluster radial thermal pressure profiles, based on
    https://ui.adsabs.harvard.edu/abs/2007ApJ...668....1N/abstract, but first used for pressure in
    https://ui.adsabs.harvard.edu/abs/2010A%26A...517A..92A/abstract.

    :param Unit/str x_unit: The unit of the x-axis of this model, kpc for instance. May be passed as a string
        representation or an astropy unit object.
    :param Unit/str y_unit: The unit of the output of this model, keV/cm^3 for instance. May be passed as a string
        representation or an astropy unit object.
    :param List[Quantity] cust_start_pars: The start values of the model parameters for any fitting function that
        used start values. The units are checked against default start values.
    """
    def __init__(self, x_unit: Union[str, Unit] = 'kpc', y_unit: Union[str, Unit] = Unit('keV cm^-3'),
                 cust_start_pars: List[Quantity] = None):
        """
        The init of a subclass of the XGA BaseModel1D class, describing a model of how thermal pressure changes
        with radius.
        """
        # If a string representation of a unit was passed then we make it an astropy unit
        if isinstance(x_unit, str):
            x_unit = Unit(x_unit)
        if isinstance(y_unit, str):
            y_unit = Unit(y_unit)

        poss_y_units = [Unit('keV cm^-3')]
        y_convertible = [u.is_equivalent(y_unit) for u in poss_y_units]
        if not any(y_convertible):
            allowed = ", ".join([u.to_string() for u in poss_y_units])
            raise UnitConversionError("{p} is not convertible to any of the allowed units; "
                                      "{a}".format(p=y_unit.to_string(), a=allowed))
        else:
            yu_ind = y_convertible.index(True)

        poss_x_units = [kpc]
        x_convertible = [u.is_equivalent(x_unit) for u in poss_x_units]
        if not any(x_convertible):
            allowed = ", ".join([u.to_string() for u in poss_x_units])
            raise UnitConversionError("{p} is not convertible to any of the allowed units; "
                                      "{a}".format(p=x_unit.to_string(), a=allowed))
        else:
            xu_ind = x_convertible.index(True)

        p_zero_starts = [Quantity(0.1, 'keV cm^-3')]
        r_scale_starts = [Quantity(100, 'kpc'), Quantity(0.2, 'deg'), Quantity(0.05, r200), Quantity(0.1, r500),
                          Quantity(0.5, r2500)]
        alpha = Quantity(1, '')
        beta = Quantity(1, '')
        gamma = Quantity(1, '')

        start_pars = [p_zero_starts[yu_ind], r_scale_starts[xu_ind], alpha, beta, gamma]

        if cust_start_pars is not None:
            # If the custom start parameters can run this gauntlet without tripping an error then we're all good
            # This method also returns the custom start pars converted to exactly the same units as the default
            start_pars = self.compare_units(cust_start_pars, start_pars)

        pz_priors = [{'prior': Quantity([0.0001, 10], 'keV cm^-3'), 'type': 'uniform'}]
        r_scale_priors = [{'prior': Quantity([0, 2000], 'kpc'), 'type': 'uniform'},
                         {'prior': Quantity([0, 1], 'deg'), 'type': 'uniform'},
                         {'prior': Quantity([0, 1], r200), 'type': 'uniform'},
                         {'prior': Quantity([0, 1], r500), 'type': 'uniform'},
                         {'prior': Quantity([0, 1], r2500), 'type': 'uniform'}]
        alpha_priors = {'prior': Quantity([0, 10], ''), 'type': 'uniform'}
        beta_priors = {'prior': Quantity([0, 10], ''), 'type': 'uniform'}
        gamma_priors = {'prior': Quantity([0, 10], ''), 'type': 'uniform'}

        priors = [pz_priors[yu_ind], r_scale_priors[xu_ind], alpha_priors, beta_priors, gamma_priors]

        nice_pars = [r"P$_{0}$", r"$R_{s}$", r"$\alpha$", r"$\beta$", r"$\gamma"]
        info_dict = {'author': 'Arnaud et al.', 'year': 2010,
                     'reference': 'https://ui.adsabs.harvard.edu/abs/2007ApJ...668....1N/abstract',
                     'general': "A model to fit galaxy cluster radial thermal pressure profiles, \n"
                                "based on the full generalized NFW profile equation."}

        super().__init__(x_unit, y_unit, start_pars, priors, 'gnfw_pressure',
                         'gNFW Pressure Profile', nice_pars, 'Thermal Pressure', info_dict)

    @staticmethod
    def model(x: Quantity, p_zero: Quantity, r_scale: Quantity, alpha: Quantity, beta: Quantity,
              gamma: Quantity) -> Quantity:
        """
        The model function for the simplified generalized NFW pressure profile.

        :param Quantity x: The radii to calculate y values for.
        :param Quantity p_zero: The pressure normalization for the model.
        :param Quantity r_scale: The scale radius for the radial pressure profile.
        :param Quantity alpha: The alpha slope parameter, for the intermediate slope.
        :param Quantity beta: The beta slope parameter, for the outer slope.
        :param Quantity gamma: The beta slope parameter, for the inner slope.
        :return: The y values corresponding to the input x values.
        :rtype: Quantity
        """
        rad_rat = x/r_scale

        result = p_zero * (1/(rad_rat**gamma * (1 + rad_rat**alpha)**((beta - gamma)/alpha)))

        return result

# So that things like fitting functions can be written generally to support different models
PRESSURE_MODELS = {"simple_gnfw_pressure": SimpleGNFWThermalPressure, "gnfw_pressure": GNFWThermalPressure}
PRESSURE_MODELS_PUB_NAMES = {n: m().publication_name for n, m in PRESSURE_MODELS.items()}
PRESSURE_MODELS_PAR_NAMES = {n: m().par_publication_names for n, m in PRESSURE_MODELS.items()}


#  This code is a part of X-ray: Generate and Analyse (XGA), a module designed for the XMM Cluster Survey (XCS).
#  Last modified by David J Turner (turne540@msu.edu) 27/04/2023, 11:02. Copyright (c) The Contributors

import os
import warnings
from typing import List, Union, Tuple

from astropy.units import Quantity, UnitConversionError

from ... import OUTPUT, NUM_CORES, XGA_EXTRACT, BASE_XSPEC_SCRIPT, XSPEC_FIT_METHOD, ABUND_TABLES
from ...samples.base import BaseSample
from ...sas import evselect_spectrum, region_setup
from ...sources import BaseSource, ExtendedSource, PointSource


def _pregen_spectra(sources: Union[BaseSource, BaseSample], outer_radius: Union[str, Quantity],
                    inner_radius: Union[str, Quantity], group_spec: bool = True, min_counts: int = 5,
                    min_sn: float = None, over_sample: float = None, one_rmf: bool = True,
                    num_cores: int = NUM_CORES) -> Tuple[Union[List[BaseSource], BaseSample], Quantity, Quantity]:
    """
    This pre-generates the spectra necessary for the requested fit (if they do not exist), and formats the input
    radii in a more predictable way.

    :param BaseSource/BaseSample sources: A single source object, or a sample of sources.
    :param str/Quantity outer_radius: The name or value of the outer radius of the region that the
        desired spectrum covers (for instance 'r200' would be acceptable for a GalaxyCluster,
        or Quantity(1000, 'kpc')). If 'region' is chosen (to use the regions in region files), then any
        inner radius will be ignored. If you are fitting for multiple sources then you can also pass a
        Quantity with one entry per source.
    :param str/Quantity inner_radius: The name or value of the outer radius of the region that the
        desired spectrum covers (for instance 'r200' would be acceptable for a GalaxyCluster,
        or Quantity(1000, 'kpc')). If 'region' is chosen (to use the regions in region files), then any
        inner radius will be ignored. By default this is zero arcseconds, resulting in a circular spectrum. If
        you are fitting for multiple sources then you can also pass a Quantity with one entry per source.
    :param bool group_spec: A boolean flag that sets whether generated spectra are grouped or not.
    :param float min_counts: If generating a grouped spectrum, this is the minimum number of counts per channel.
        To disable minimum counts set this parameter to None.
    :param float min_sn: If generating a grouped spectrum, this is the minimum signal to noise in each channel.
        To disable minimum signal to noise set this parameter to None.
    :param float over_sample: The minimum energy resolution for each group, set to None to disable. e.g. if
        over_sample=3 then the minimum width of a group is 1/3 of the resolution FWHM at that energy.
    :param bool one_rmf: This flag tells the method whether it should only generate one RMF for a particular
        ObsID-instrument combination - this is much faster in some circumstances, however the RMF does depend
        slightly on position on the detector.
    :param int num_cores: The number of cores to use (if running locally), default is set to 90% of available.
    :return: Most likely just the passed in sources, but if a single source was passed
    then a list will be returned.
    :rtype: Union[List[BaseSource], BaseSample]
    """
    # I call the evselect_spectrum function here for two reasons; to make sure that the spectra which the user
    #  want to fit are generated, and because that function has a lot of radius parsing and checking stuff
    #  in it which will kick up a fuss if variables aren't formatted right
    sources = evselect_spectrum(sources, outer_radius, inner_radius, group_spec, min_counts, min_sn, over_sample,
                                one_rmf, num_cores)

    # This is the spectrum region preparation function, and I'm calling it here because it will return properly
    #  formatted arrays for the inner and outer radii
    if outer_radius != 'region':
        inn_rad_vals, out_rad_vals = region_setup(sources, outer_radius, inner_radius, True, '')[1:]
    else:
        raise NotImplementedError("I don't currently support fitting region spectra")

    return sources, inn_rad_vals, out_rad_vals


def _check_inputs(sources: Union[BaseSource, BaseSample], lum_en: Quantity, lo_en: Quantity, hi_en: Quantity,
                  fit_method: str, abund_table: str, timeout: Quantity) -> Union[List[BaseSource], BaseSample]:
    """
    This performs some checks that are common to all the model fit functions, also makes sure the necessary spectra
    have been generated.

    :param BaseSource/BaseSample sources: A single source object, or a sample of sources.
    :param Quantity lum_en: Energy bands in which to measure luminosity.
    :param Quantity lo_en: The lower energy limit for the data to be fitted.
    :param Quantity hi_en: The upper energy limit for the data to be fitted.
    :param str fit_method: The XSPEC fit method to use.
    :param str abund_table: The abundance table to use for the fit.
    :param Quantity timeout: The amount of time each individual fit is allowed to run for, the default is one hour.
        Please note that this is not a timeout for the entire fitting process, but a timeout to individual source
        fits.
    :return: Most likely just the passed in sources, but if a single source was passed
    then a list will be returned.
    :rtype: Union[List[BaseSource], BaseSample]
    """
    # This function supports passing both individual sources and samples of sources, but I do require that
    #  the sources object is iterable
    if isinstance(sources, BaseSource):
        sources = [sources]

    # Not allowed to use BaseSources for this, though they shouldn't have spectra anyway
    if not all([isinstance(src, (ExtendedSource, PointSource)) for src in sources]):
        raise TypeError("This convenience function cannot be used with BaseSource objects.")
    elif not all([src.detected for src in sources]):
        warnings.warn("Not all of these sources have been detected, you may get a poor fit.")

    # Checks that the luminosity energy bands are pairs of values
    if lum_en.shape[1] != 2:
        raise ValueError("Luminosity energy bands should be supplied in pairs, defined "
                         "like Quantity([[0.5, 2.0], [2.0, 10.0]], 'keV')")
    # Are the lower limits smaller than the upper limits? - Obviously they should be so I check
    elif not all([lum_en[pair_ind, 0] < lum_en[pair_ind, 1] for pair_ind in range(0, lum_en.shape[0])]):
        raise ValueError("Luminosity energy band first entries must be smaller than second entries.")

    # These are different energy limits to those above, these are what govern how much of the data we fit to.
    # Do the same check to make sure lower limit is less than upper limit
    if lo_en > hi_en:
        raise ValueError("lo_en cannot be greater than hi_en.")

    if fit_method not in XSPEC_FIT_METHOD:
        raise ValueError("{f} is not an XSPEC fit method, allowed fit methods are; "
                         "{a}.".format(f=fit_method, a=", ".join(XSPEC_FIT_METHOD)))

    if abund_table not in ABUND_TABLES:
        raise ValueError("{f} is not an XSPEC abundance table, allowed abundance tables are; "
                         "{a}.".format(f=fit_method, a=", ".join(ABUND_TABLES)))

    if not timeout.unit.is_equivalent('second'):
        raise UnitConversionError("The timeout quantity must be in units which can be converted to seconds, you have"
                                  " passed a quantity with units of {}".format(timeout.unit.to_string()))

    return sources


def _write_xspec_script(source: BaseSource, spec_storage_key: str, model: str, abund_table: str, fit_method: str,
                        specs: str, lo_en: Quantity, hi_en: Quantity, par_names: str, par_values: str,
                        linking: str, freezing: str, par_fit_stat: float, lum_low_lims: str, lum_upp_lims: str,
                        lum_conf: float, redshift: float, pre_check: bool, check_par_names: str, check_par_lo_lims: str,
                        check_par_hi_lims: str, check_par_err_lims: str, norm_scale: bool,
                        which_par_nh: str = 'None') -> Tuple[str, str]:
    """
    This writes out a configured XSPEC script, and is common to all fit functions.

    :param BaseSource source: The source for which an XSPEC script is being created
    :param str spec_storage_key: The storage key that the spectra that have been included in the current fit
        are stored under.
    :param str model: The model being fitted to the data.
    :param str abund_table: The chosen abundance table for XSPEC to use.
    :param str fit_method: Which fit method should XSPEC use to fit the model to data.
    :param str specs: A string containing the paths to all spectra to be fitted.
    :param Quantity lo_en: The lower energy limit for the data to be fitted.
    :param Quantity hi_en: The upper energy limit for the data to be fitted.
    :param str par_names: A string containing the names of the model parameters.
    :param str par_values: A string containing the start values of the model parameters.
    :param str linking: A string containing the linking settings for the model.
    :param str freezing: A string containing the freezing settings for the model.
    :param float par_fit_stat: The delta fit statistic for the XSPEC 'error' command.
    :param str lum_low_lims: A string containing the lower energy limits for the luminosity measurements.
    :param str lum_upp_lims: A string containing the upper energy limits for the luminosity measurements.
    :param float lum_conf: The confidence level for XSPEC luminosity measurements.
    :param float redshift: The redshift of the object.
    :param bool pre_check: Flag indicating whether a pre-check of the quality of the input spectra
        should be performed.
    :param str check_par_names: A string representing a TCL list of model parameter names that checks should be
        performed on.
    :param str check_par_lo_lims: A string representing a TCL list of allowed lower limits for the check_par_names
        parameter entries.
    :param str check_par_hi_lims: A string representing a TCL list of allowed upper limits for the check_par_names
        parameter entries.
    :param str check_par_err_lims: A string representing a TCL list of allowed upper limits for the parameter
        uncertainties.
    :param bool norm_scale: Is there an extra constant designed to account for the differences in normalisation
        you can get from different observations of a cluster.
    :param str which_par_nh: The parameter IDs of the nH parameters values which should be zeroed for the calculation
        of unabsorbed luminosities.
    :return: The paths to the output file and the script file.
    :rtype: Tuple[str, str]
    """
    # Read in the template file for the XSPEC script.
    with open(BASE_XSPEC_SCRIPT, 'r') as x_script:
        script = x_script.read()

    # There has to be a directory to write this xspec script to, as well as somewhere for the fit output
    #  to be stored
    dest_dir = OUTPUT + "XSPEC/" + source.name + "/"
    if not os.path.exists(dest_dir):
        os.makedirs(dest_dir)
    # Defining where the output summary file of the fit is written
    out_file = dest_dir + source.name + "_" + spec_storage_key + "_" + model
    script_file = dest_dir + source.name + "_" + spec_storage_key + "_" + model + ".xcm"

    # The template is filled out here, taking everything we have generated and everything the user
    #  passed in. The result is an XSPEC script that can be run as is.
    script = script.format(xsp=XGA_EXTRACT, ab=abund_table, md=fit_method, H0=source.cosmo.H0.value,
                           q0=0., lamb0=source.cosmo.Ode0, sp=specs, lo_cut=lo_en.to("keV").value,
                           hi_cut=hi_en.to("keV").value, m=model, pn=par_names, pv=par_values,
                           lk=linking, fr=freezing, el=par_fit_stat, lll=lum_low_lims, lul=lum_upp_lims,
                           of=out_file, redshift=redshift, lel=lum_conf, check=pre_check, cps=check_par_names,
                           cpsl=check_par_lo_lims, cpsh=check_par_hi_lims, cpse=check_par_err_lims, ns=norm_scale,
                           nhmtz=which_par_nh)

    # Write out the filled-in template to its destination
    with open(script_file, 'w') as xcm:
        xcm.write(script)

    return out_file, script_file
